# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from inspect import signature
from itertools import chain
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)
from numpy import ndarray
import dearpygui.dearpygui as dpg
import deareis.themes as themes
from deareis.tooltips import attach_tooltip
import deareis.tooltips as tooltips
from deareis.enums import (
    Context,
    DRTMethod,
    PlotType,
    label_to_plot_type,
    plot_type_to_label,
)
from deareis.data import (
    DRTResult,
    DataSet,
    FitResult,
    PlotSettings,
    SimulationResult,
    TestResult,
)
from deareis.gui.plots import (
    BodeMagnitude,
    BodePhase,
    DRT,
    ImpedanceImaginary,
    ImpedanceReal,
    Nyquist,
)
from deareis.gui.plots.base import Plot
from deareis.signals import Signal
import deareis.signals as signals
from deareis.utility import is_filtered_item_visible


TABLE_HEADER_HEIGHT: int = 18
TABLE_ROW_HEIGHT: int = 23


class DataSetsGroup:
    def __init__(self):
        self.header: int = dpg.generate_uuid()
        with dpg.collapsing_header(
            label="Data sets",
            tag=self.header,
        ):
            with dpg.group(horizontal=True, indent=8):
                self.select_all_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Select all",
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, **u
                    ),
                    user_data={},
                    tag=self.select_all_button,
                )
                self.unselect_all_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Unselect all",
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, **u
                    ),
                    user_data={},
                    tag=self.unselect_all_button,
                )
            self.table: int = dpg.generate_uuid()
            with dpg.group(indent=8):
                with dpg.table(
                    borders_outerV=True,
                    borders_outerH=True,
                    borders_innerV=True,
                    borders_innerH=True,
                    scrollY=True,
                    freeze_rows=1,
                    width=-1,
                    height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT,
                    tag=self.table,
                ):
                    dpg.add_table_column(label="", width_fixed=True)
                    attach_tooltip(tooltips.plotting.dataset_checkbox)
                    dpg.add_table_column(label="Label", width_fixed=True)
                    attach_tooltip(tooltips.plotting.dataset_label)
            dpg.add_spacer(height=8)
        self.data_hash: str = ""
        self.active_hash: str = ""

    def clear(self):
        dpg.delete_item(self.table, children_only=True, slot=1)
        dpg.configure_item(self.table, height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT)
        dpg.hide_item(self.header)
        self.data_hash = ""
        self.active_hash = ""

    def populate(self, data_sets: List[DataSet], settings: PlotSettings) -> bool:
        assert type(data_sets) is list, data_sets
        assert type(settings) is PlotSettings, settings
        data_hash: str = ",".join([_.uuid for _ in data_sets])
        if data_hash == self.data_hash:
            return False
        self.clear()
        self.data_hash = data_hash
        if not data_sets:
            dpg.configure_item(
                self.select_all_button,
                user_data={},
            )
            dpg.configure_item(
                self.unselect_all_button,
                user_data={},
            )
            return True
        data: DataSet
        for data in data_sets:
            with dpg.table_row(
                filter_key=data.get_label().lower(),
                parent=self.table,
            ):
                dpg.add_checkbox(
                    default_value=data.uuid in settings.series_order,
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, enabled=a, **u
                    ),
                    user_data={
                        "data_sets": [data],
                        "settings": settings,
                    },
                )
                dpg.add_text(data.get_label())
                attach_tooltip(data.get_label())
        dpg.configure_item(
            self.table,
            height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT * max(1, len(data_sets)),
        )
        dpg.show_item(self.header)
        return True

    def update(
        self,
        active_hash: str,
        data_sets: List[DataSet],
        settings: PlotSettings,
    ):
        assert type(active_hash) is str, active_hash
        assert type(data_sets) is list, data_sets
        assert type(settings) is PlotSettings, settings
        if active_hash == self.active_hash:
            return
        self.active_hash = active_hash
        data: DataSet
        row: int
        for (data, row) in zip(data_sets, dpg.get_item_children(self.table, slot=1)):
            cells: List[int] = dpg.get_item_children(row, slot=1)
            dpg.configure_item(
                cells[0],
                default_value=data.uuid in settings.series_order,
                user_data={
                    "data_sets": [data],
                    "settings": settings,
                },
            )

    def filter(self, string: str, collapse: bool) -> List[DataSet]:
        assert type(string) is str, string
        stripped_string: str = string.strip()
        data_sets: List[DataSet] = []
        dpg.set_value(self.table, string)
        row: int
        for row in dpg.get_item_children(self.table, slot=1):
            filter_key: str = dpg.get_item_filter_key(row)
            subset: List[DataSet] = dpg.get_item_user_data(
                dpg.get_item_children(row, slot=1)[0]
            ).get("data_sets", [])
            if is_filtered_item_visible(row, stripped_string):
                data_sets.extend(subset)
        dpg.configure_item(
            self.table,
            height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT * len(data_sets),
        )
        if data_sets:
            dpg.show_item(self.header)
            if collapse:
                dpg.set_value(self.header, not string == "")
        else:
            if collapse:
                dpg.set_value(self.header, False)
            dpg.hide_item(self.header)
        dpg.get_item_user_data(self.select_all_button).update({"data_sets": data_sets})
        dpg.get_item_user_data(self.unselect_all_button).update(
            {"data_sets": data_sets}
        )
        dpg.set_item_label(self.header, f"Data sets ({len(data_sets)})")
        return data_sets


class TestResultGroup:
    def __init__(self, parent: int):
        self.header: int = dpg.generate_uuid()
        with dpg.collapsing_header(
            label="PLACEHOLDER", show=False, tag=self.header, parent=parent, indent=8
        ):
            with dpg.group(horizontal=True, indent=8):
                self.select_all_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Select all",
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, **u
                    ),
                    user_data={},
                    tag=self.select_all_button,
                )
                self.unselect_all_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Unselect all",
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, **u
                    ),
                    user_data={},
                    tag=self.unselect_all_button,
                )
            with dpg.group(indent=8):
                self.table: int = dpg.generate_uuid()
                with dpg.table(
                    borders_outerV=True,
                    borders_outerH=True,
                    borders_innerV=True,
                    borders_innerH=True,
                    scrollY=True,
                    freeze_rows=1,
                    width=-1,
                    height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT,
                    tag=self.table,
                ):
                    dpg.add_table_column(label="", width_fixed=True)
                    attach_tooltip(tooltips.plotting.test_checkbox)
                    dpg.add_table_column(label="Label", width_fixed=True)
            dpg.add_spacer(height=8)
        self.test_hash: str = ""
        self.active_hash: str = ""

    def clear(self):
        dpg.delete_item(self.table, children_only=True, slot=1)
        dpg.configure_item(self.table, height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT)
        dpg.hide_item(self.header)
        self.test_hash = ""
        self.active_hash = ""

    def populate(
        self, tests: List[TestResult], data: DataSet, settings: PlotSettings
    ) -> bool:
        assert type(tests) is list, tests
        assert type(data) is DataSet, data
        assert type(settings) is PlotSettings, settings
        test_hash: str = ",".join([_.uuid for _ in tests])
        if test_hash == self.test_hash:
            return False
        self.clear()
        self.test_hash = test_hash
        dpg.set_item_label(self.header, data.get_label())
        test: TestResult
        for test in tests:
            with dpg.table_row(
                filter_key=f"{data.get_label().lower()} {test.get_label().lower()}",
                parent=self.table,
            ):
                dpg.add_checkbox(
                    default_value=test.uuid in settings.series_order,
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, enabled=a, **u
                    ),
                    user_data={
                        "tests": [test],
                        "settings": settings,
                    },
                )
                dpg.add_text(test.get_label())
                attach_tooltip(test.get_label())
        dpg.configure_item(
            self.table,
            height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT * max(1, len(tests)),
        )
        dpg.show_item(self.header)
        return True

    def update(
        self,
        active_hash: str,
        tests: List[TestResult],
        data: DataSet,
        settings: PlotSettings,
    ):
        assert type(active_hash) is str, active_hash
        assert type(tests) is list, tests
        assert type(data) is DataSet, data
        assert type(settings) is PlotSettings, settings
        if active_hash == self.active_hash:
            return
        self.active_hash = active_hash
        test: TestResult
        row: int
        for (test, row) in zip(tests, dpg.get_item_children(self.table, slot=1)):
            cells: List[int] = dpg.get_item_children(row, slot=1)
            dpg.configure_item(
                cells[0],
                default_value=test.uuid in settings.series_order,
                user_data={
                    "tests": [test],
                    "settings": settings,
                },
            )

    def filter(self, string: str, collapse: bool) -> List[TestResult]:
        assert type(string) is str, string
        stripped_string: str = string.strip()
        dpg.set_value(self.table, string)
        tests: List[TestResult] = []
        row: int
        for row in dpg.get_item_children(self.table, slot=1):
            filter_key: str = dpg.get_item_filter_key(row)
            subset: List[DataSet] = dpg.get_item_user_data(
                dpg.get_item_children(row, slot=1)[0]
            ).get("tests", [])
            if is_filtered_item_visible(row, stripped_string):
                tests.extend(subset)
        dpg.configure_item(
            self.table, height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT * len(tests)
        )
        if tests:
            dpg.show_item(self.header)
            if collapse:
                dpg.set_value(self.header, not string == "")
        else:
            if collapse:
                dpg.set_value(self.header, False)
            dpg.hide_item(self.header)
        dpg.get_item_user_data(self.select_all_button).update({"tests": tests})
        dpg.get_item_user_data(self.unselect_all_button).update({"tests": tests})
        return tests


class TestsGroup:
    def __init__(self):
        self.groups: Dict[str, TestResultGroup] = {}
        self.header: int = dpg.generate_uuid()
        with dpg.collapsing_header(
            label="Kramers-Kronig tests",
            tag=self.header,
        ):
            self.button_group: int = dpg.generate_uuid()
            with dpg.group(horizontal=True, indent=8, tag=self.button_group):
                dpg.add_button(
                    label="Expand all",
                    callback=lambda s, a, u: self.expand_subheaders(True),
                )
                dpg.add_button(
                    label="Collapse all",
                    callback=lambda s, a, u: self.expand_subheaders(False),
                )
                self.select_all_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Select all",
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, **u
                    ),
                    user_data={},
                    tag=self.select_all_button,
                )
                self.unselect_all_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Unselect all",
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, **u
                    ),
                    user_data={},
                    tag=self.unselect_all_button,
                )
            dpg.add_spacer(height=8)
        self.data_hash: str = ""

    def expand_subheaders(self, state: bool):
        assert type(state) is bool
        subheader: int
        for subheader in dpg.get_item_children(self.header, slot=1):
            if "::mvCollapsingHeader" not in dpg.get_item_type(subheader):
                continue
            dpg.set_value(subheader, state)

    def clear(self):
        group: TestResultGroup
        for group in self.groups.values():
            group.clear()
        dpg.hide_item(self.header)
        self.data_hash = ""

    def populate(
        self,
        tests: Dict[str, List[TestResult]],
        data_sets: List[DataSet],
        settings: PlotSettings,
    ):
        assert type(tests) is dict, tests
        assert type(data_sets) is list, data_sets
        assert type(settings) is PlotSettings, settings
        if not data_sets:
            self.clear()
            dpg.configure_item(
                self.select_all_button,
                user_data={},
            )
            dpg.configure_item(
                self.unselect_all_button,
                user_data={},
            )
            return
        data_hash: str = ",".join([_.uuid for _ in data_sets])
        if data_hash != self.data_hash:
            self.clear()
            self.data_hash = data_hash
        active_hash: str = ",".join(sorted(settings.series_order)) + f",{settings.uuid}"
        all_tests: List[TestResult] = []
        data: DataSet
        for data in data_sets:
            if not tests[data.uuid]:
                if data.uuid in self.groups:
                    self.groups[data.uuid].clear()
                continue
            if data.uuid not in self.groups:
                self.groups[data.uuid] = TestResultGroup(self.header)
            group: TestResultGroup = self.groups[data.uuid]
            if not group.populate(tests[data.uuid], data, settings):
                group.update(active_hash, tests[data.uuid], data, settings)
            dpg.get_item_user_data(group.select_all_button).update(
                {
                    "enabled": True,
                    "tests": tests,
                    "settings": settings,
                }
            )
            dpg.get_item_user_data(group.unselect_all_button).update(
                {
                    "enabled": False,
                    "tests": tests,
                    "settings": settings,
                }
            )
            all_tests.extend(tests[data.uuid])
        if all_tests:
            dpg.show_item(self.header)
        dpg.get_item_user_data(self.select_all_button).update(
            {
                "enabled": True,
                "tests": all_tests,
                "settings": settings,
            }
        )
        dpg.get_item_user_data(self.unselect_all_button).update(
            {
                "enabled": False,
                "tests": all_tests,
                "settings": settings,
            }
        )

    def filter(self, string: str, collapse: bool) -> List[TestResult]:
        assert type(string) is str, string
        tests: List[TestResult] = []
        group: TestResultGroup
        for group in self.groups.values():
            tests.extend(group.filter(string, collapse))
        if tests:
            dpg.show_item(self.header)
            if collapse:
                dpg.set_value(self.header, not string == "")
        else:
            if collapse:
                dpg.set_value(self.header, False)
            dpg.hide_item(self.header)
        dpg.get_item_user_data(self.select_all_button).update({"tests": tests})
        dpg.get_item_user_data(self.unselect_all_button).update({"tests": tests})
        dpg.set_item_label(self.header, f"Kramers-Kronig tests ({len(tests)})")
        return tests


class DRTResultGroup:
    def __init__(self, parent: int):
        self.header: int = dpg.generate_uuid()
        with dpg.collapsing_header(
            label="PLACEHOLDER", show=False, tag=self.header, parent=parent, indent=8
        ):
            with dpg.group(horizontal=True, indent=8):
                self.select_all_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Select all",
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, **u
                    ),
                    user_data={},
                    tag=self.select_all_button,
                )
                self.unselect_all_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Unselect all",
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, **u
                    ),
                    user_data={},
                    tag=self.unselect_all_button,
                )
            with dpg.group(indent=8):
                self.table: int = dpg.generate_uuid()
                with dpg.table(
                    borders_outerV=True,
                    borders_outerH=True,
                    borders_innerV=True,
                    borders_innerH=True,
                    scrollY=True,
                    freeze_rows=1,
                    width=-1,
                    height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT,
                    tag=self.table,
                ):
                    dpg.add_table_column(label="", width_fixed=True)
                    attach_tooltip(tooltips.plotting.drt_checkbox)
                    dpg.add_table_column(label="Label", width_fixed=True)
            dpg.add_spacer(height=8)
        self.drt_hash: str = ""
        self.active_hash: str = ""

    def clear(self):
        dpg.delete_item(self.table, children_only=True, slot=1)
        dpg.configure_item(self.table, height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT)
        dpg.hide_item(self.header)
        self.drt_hash = ""
        self.active_hash = ""

    def populate(
        self,
        drts: List[DRTResult],
        data: DataSet,
        settings: PlotSettings,
    ) -> bool:
        assert type(drts) is list, drts
        assert type(data) is DataSet, data
        assert type(settings) is PlotSettings, settings
        drt_hash: str = ",".join([_.uuid for _ in drts])
        if drt_hash == self.drt_hash:
            return False
        self.clear()
        self.drt_hash = drt_hash
        dpg.set_item_label(self.header, data.get_label())
        drt: DRTResult
        for drt in drts:
            with dpg.table_row(
                filter_key=f"{data.get_label().lower()} {drt.get_label().lower()}",
                parent=self.table,
            ):
                dpg.add_checkbox(
                    default_value=drt.uuid in settings.series_order,
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, enabled=a, **u
                    ),
                    user_data={
                        "drts": [drt],
                        "settings": settings,
                    },
                )
                dpg.add_text(drt.get_label())
                attach_tooltip(drt.get_label())
        dpg.configure_item(
            self.table,
            height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT * max(1, len(drts)),
        )
        dpg.show_item(self.header)
        return True

    def update(
        self,
        active_hash: str,
        drts: List[DRTResult],
        data: DataSet,
        settings: PlotSettings,
    ):
        assert type(active_hash) is str, active_hash
        assert type(drts) is list, drts
        assert type(data) is DataSet, data
        assert type(settings) is PlotSettings, settings
        if active_hash == self.active_hash:
            return
        self.active_hash = active_hash
        drt: DRTResult
        row: int
        for (drt, row) in zip(drts, dpg.get_item_children(self.table, slot=1)):
            cells: List[int] = dpg.get_item_children(row, slot=1)
            dpg.configure_item(
                cells[0],
                default_value=drt.uuid in settings.series_order,
                user_data={
                    "drts": [drt],
                    "settings": settings,
                },
            )

    def filter(self, string: str, collapse: bool) -> List[DRTResult]:
        assert type(string) is str, string
        stripped_string: str = string.strip()
        dpg.set_value(self.table, string)
        drts: List[DRTResult] = []
        row: int
        for row in dpg.get_item_children(self.table, slot=1):
            filter_key: str = dpg.get_item_filter_key(row)
            subset: List[DataSet] = dpg.get_item_user_data(
                dpg.get_item_children(row, slot=1)[0]
            ).get("drts", [])
            if is_filtered_item_visible(row, stripped_string):
                drts.extend(subset)
        dpg.configure_item(
            self.table, height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT * len(drts)
        )
        if drts:
            dpg.show_item(self.header)
            if collapse:
                dpg.set_value(self.header, not string == "")
        else:
            if collapse:
                dpg.set_value(self.header, False)
            dpg.hide_item(self.header)
        dpg.get_item_user_data(self.select_all_button).update({"drts": drts})
        dpg.get_item_user_data(self.unselect_all_button).update({"drts": drts})
        return drts


class DRTsGroup:
    def __init__(self):
        self.groups: Dict[str, DRTResultGroup] = {}
        self.header: int = dpg.generate_uuid()
        with dpg.collapsing_header(
            label="DRT analyses",
            tag=self.header,
        ):
            self.button_group: int = dpg.generate_uuid()
            with dpg.group(horizontal=True, indent=8, tag=self.button_group):
                dpg.add_button(
                    label="Expand all",
                    callback=lambda s, a, u: self.expand_subheaders(True),
                )
                dpg.add_button(
                    label="Collapse all",
                    callback=lambda s, a, u: self.expand_subheaders(False),
                )
                self.select_all_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Select all",
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, **u
                    ),
                    user_data={},
                    tag=self.select_all_button,
                )
                self.unselect_all_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Unselect all",
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, **u
                    ),
                    user_data={},
                    tag=self.unselect_all_button,
                )
            dpg.add_spacer(height=8)
        self.data_hash: str = ""

    def expand_subheaders(self, state: bool):
        assert type(state) is bool
        subheader: int
        for subheader in dpg.get_item_children(self.header, slot=1):
            if "::mvCollapsingHeader" not in dpg.get_item_type(subheader):
                continue
            dpg.set_value(subheader, state)

    def clear(self):
        group: DRTResultGroup
        for group in self.groups.values():
            group.clear()
        dpg.hide_item(self.header)
        self.data_hash = ""

    def populate(
        self,
        drts: Dict[str, List[DRTResult]],
        data_sets: List[DataSet],
        settings: PlotSettings,
    ):
        assert type(drts) is dict, drts
        assert type(data_sets) is list, data_sets
        assert type(settings) is PlotSettings, settings
        if not data_sets:
            self.clear()
            dpg.configure_item(
                self.select_all_button,
                user_data={},
            )
            dpg.configure_item(
                self.unselect_all_button,
                user_data={},
            )
            return
        data_hash: str = ",".join([_.uuid for _ in data_sets])
        if data_hash != self.data_hash:
            self.clear()
            self.data_hash = data_hash
        active_hash: str = ",".join(sorted(settings.series_order)) + f",{settings.uuid}"
        all_drts: List[DRTResult] = []
        data: DataSet
        for data in data_sets:
            if not drts[data.uuid]:
                if data.uuid in self.groups:
                    self.groups[data.uuid].clear()
                continue
            if data.uuid not in self.groups:
                self.groups[data.uuid] = DRTResultGroup(self.header)
            group: DRTResultGroup = self.groups[data.uuid]
            if not group.populate(drts[data.uuid], data, settings):
                group.update(active_hash, drts[data.uuid], data, settings)
            dpg.get_item_user_data(group.select_all_button).update(
                {
                    "enabled": True,
                    "drts": drts,
                    "settings": settings,
                }
            )
            dpg.get_item_user_data(group.unselect_all_button).update(
                {
                    "enabled": False,
                    "drts": drts,
                    "settings": settings,
                }
            )
            all_drts.extend(drts[data.uuid])
        if all_drts:
            dpg.show_item(self.header)
        dpg.get_item_user_data(self.select_all_button).update(
            {
                "enabled": True,
                "drts": all_drts,
                "settings": settings,
            }
        )
        dpg.get_item_user_data(self.unselect_all_button).update(
            {
                "enabled": False,
                "drts": all_drts,
                "settings": settings,
            }
        )

    def filter(self, string: str, collapse: bool) -> List[DRTResult]:
        assert type(string) is str, string
        drts: List[DRTResult] = []
        group: DRTResultGroup
        for group in self.groups.values():
            drts.extend(group.filter(string, collapse))
        if drts:
            dpg.show_item(self.header)
            if collapse:
                dpg.set_value(self.header, not string == "")
        else:
            if collapse:
                dpg.set_value(self.header, False)
            dpg.hide_item(self.header)
        dpg.get_item_user_data(self.select_all_button).update({"drts": drts})
        dpg.get_item_user_data(self.unselect_all_button).update({"drts": drts})
        dpg.set_item_label(self.header, f"DRT analyses ({len(drts)})")
        return drts


class FitResultGroup:
    def __init__(self, parent: int):
        self.header: int = dpg.generate_uuid()
        with dpg.collapsing_header(
            label="PLACEHOLDER", show=False, tag=self.header, parent=parent, indent=8
        ):
            with dpg.group(horizontal=True, indent=8):
                self.select_all_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Select all",
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, **u
                    ),
                    user_data={},
                    tag=self.select_all_button,
                )
                self.unselect_all_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Unselect all",
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, **u
                    ),
                    user_data={},
                    tag=self.unselect_all_button,
                )
            with dpg.group(indent=8):
                self.table: int = dpg.generate_uuid()
                with dpg.table(
                    borders_outerV=True,
                    borders_outerH=True,
                    borders_innerV=True,
                    borders_innerH=True,
                    scrollY=True,
                    freeze_rows=1,
                    width=-1,
                    height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT,
                    tag=self.table,
                ):
                    dpg.add_table_column(label="", width_fixed=True)
                    attach_tooltip(tooltips.plotting.fit_checkbox)
                    dpg.add_table_column(label="Label", width_fixed=True)
            dpg.add_spacer(height=8)
        self.fit_hash: str = ""
        self.active_hash: str = ""

    def clear(self):
        dpg.delete_item(self.table, children_only=True, slot=1)
        dpg.configure_item(self.table, height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT)
        dpg.hide_item(self.header)
        self.fit_hash = ""
        self.active_hash = ""

    def populate(
        self,
        fits: List[FitResult],
        data: DataSet,
        settings: PlotSettings,
    ) -> bool:
        assert type(fits) is list, fits
        assert type(data) is DataSet, data
        assert type(settings) is PlotSettings, settings
        fit_hash: str = ",".join([_.uuid for _ in fits])
        if fit_hash == self.fit_hash:
            return False
        self.clear()
        self.fit_hash = fit_hash
        dpg.set_item_label(self.header, data.get_label())
        fit: FitResult
        for fit in fits:
            with dpg.table_row(
                filter_key=f"{data.get_label().lower()} {fit.get_label().lower()}",
                parent=self.table,
            ):
                dpg.add_checkbox(
                    default_value=fit.uuid in settings.series_order,
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, enabled=a, **u
                    ),
                    user_data={
                        "fits": [fit],
                        "settings": settings,
                    },
                )
                dpg.add_text(fit.get_label())
                attach_tooltip(fit.get_label())
        dpg.configure_item(
            self.table,
            height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT * max(1, len(fits)),
        )
        dpg.show_item(self.header)
        return True

    def update(
        self,
        active_hash: str,
        fits: List[FitResult],
        data: DataSet,
        settings: PlotSettings,
    ):
        assert type(active_hash) is str, active_hash
        assert type(fits) is list, fits
        assert type(data) is DataSet, data
        assert type(settings) is PlotSettings, settings
        if active_hash == self.active_hash:
            return
        self.active_hash = active_hash
        fit: FitResult
        row: int
        for (fit, row) in zip(fits, dpg.get_item_children(self.table, slot=1)):
            cells: List[int] = dpg.get_item_children(row, slot=1)
            dpg.configure_item(
                cells[0],
                default_value=fit.uuid in settings.series_order,
                user_data={
                    "fits": [fit],
                    "settings": settings,
                },
            )

    def filter(self, string: str, collapse: bool) -> List[FitResult]:
        assert type(string) is str, string
        stripped_string: str = string.strip()
        dpg.set_value(self.table, string)
        fits: List[FitResult] = []
        row: int
        for row in dpg.get_item_children(self.table, slot=1):
            filter_key: str = dpg.get_item_filter_key(row)
            subset: List[DataSet] = dpg.get_item_user_data(
                dpg.get_item_children(row, slot=1)[0]
            ).get("fits", [])
            if is_filtered_item_visible(row, stripped_string):
                fits.extend(subset)
        dpg.configure_item(
            self.table,
            height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT * max(1, len(fits)),
        )
        if fits:
            dpg.show_item(self.header)
            if collapse:
                dpg.set_value(self.header, not string == "")
        else:
            if collapse:
                dpg.set_value(self.header, False)
            dpg.hide_item(self.header)
        dpg.get_item_user_data(self.select_all_button).update({"fits": fits})
        dpg.get_item_user_data(self.unselect_all_button).update({"fits": fits})
        return fits


class FitsGroup:
    def __init__(self):
        self.groups: Dict[str, FitResultGroup] = {}
        self.header: int = dpg.generate_uuid()
        with dpg.collapsing_header(
            label="Fitted equivalent circuits",
            tag=self.header,
        ):
            self.button_group: int = dpg.generate_uuid()
            with dpg.group(horizontal=True, indent=8, tag=self.button_group):
                dpg.add_button(
                    label="Expand all",
                    callback=lambda s, a, u: self.expand_subheaders(True),
                )
                dpg.add_button(
                    label="Collapse all",
                    callback=lambda s, a, u: self.expand_subheaders(False),
                )
                self.select_all_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Select all",
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, **u
                    ),
                    user_data={},
                    tag=self.select_all_button,
                )
                self.unselect_all_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Unselect all",
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, **u
                    ),
                    user_data={},
                    tag=self.unselect_all_button,
                )
            dpg.add_spacer(height=8)
        self.data_hash: str = ""

    def expand_subheaders(self, state: bool):
        assert type(state) is bool
        subheader: int
        for subheader in dpg.get_item_children(self.header, slot=1):
            if "::mvCollapsingHeader" not in dpg.get_item_type(subheader):
                continue
            dpg.set_value(subheader, state)

    def clear(self):
        group: FitResultGroup
        for group in self.groups.values():
            group.clear()
        dpg.hide_item(self.header)
        self.data_hash = ""

    def populate(
        self,
        fits: Dict[str, List[FitResult]],
        data_sets: List[DataSet],
        settings: PlotSettings,
    ):
        assert type(fits) is dict, fits
        assert type(data_sets) is list, data_sets
        assert type(settings) is PlotSettings, settings
        if not data_sets:
            self.clear()
            dpg.configure_item(
                self.select_all_button,
                user_data={},
            )
            dpg.configure_item(
                self.unselect_all_button,
                user_data={},
            )
            return
        data_hash: str = ",".join([_.uuid for _ in data_sets])
        if data_hash != self.data_hash:
            self.clear()
            self.data_hash = data_hash
        active_hash: str = ",".join(sorted(settings.series_order)) + f",{settings.uuid}"
        all_fits: List[FitResult] = []
        data: DataSet
        for data in data_sets:
            if not fits[data.uuid]:
                if data.uuid in self.groups:
                    self.groups[data.uuid].clear()
                continue
            if data.uuid not in self.groups:
                self.groups[data.uuid] = FitResultGroup(self.header)
            group: FitResultGroup = self.groups[data.uuid]
            if not group.populate(fits[data.uuid], data, settings):
                group.update(active_hash, fits[data.uuid], data, settings)
            dpg.get_item_user_data(group.select_all_button).update(
                {
                    "enabled": True,
                    "fits": fits[data.uuid],
                    "settings": settings,
                }
            )
            dpg.get_item_user_data(group.unselect_all_button).update(
                {
                    "enabled": False,
                    "fits": fits[data.uuid],
                    "settings": settings,
                }
            )
            all_fits.extend(fits[data.uuid])
        if all_fits:
            dpg.show_item(self.header)
        dpg.get_item_user_data(self.select_all_button).update(
            {
                "enabled": True,
                "fits": all_fits,
                "settings": settings,
            }
        )
        dpg.get_item_user_data(self.unselect_all_button).update(
            {
                "enabled": False,
                "fits": all_fits,
                "settings": settings,
            }
        )

    def filter(self, string: str, collapse: bool) -> List[FitResult]:
        assert type(string) is str, string
        # TODO: Determine whether to filter
        fits: List[FitResult] = []
        group: FitResultGroup
        for group in self.groups.values():
            fits.extend(group.filter(string, collapse))
        if fits:
            dpg.show_item(self.header)
            if collapse:
                dpg.set_value(self.header, not string == "")
        else:
            if collapse:
                dpg.set_value(self.header, False)
            dpg.hide_item(self.header)
        dpg.get_item_user_data(self.select_all_button).update({"fits": fits})
        dpg.get_item_user_data(self.unselect_all_button).update({"fits": fits})
        dpg.set_item_label(self.header, f"Fitted equivalent circuits ({len(fits)})")
        return fits


class SimulationsGroup:
    def __init__(self):
        self.header: int = dpg.generate_uuid()
        with dpg.collapsing_header(
            label="Simulated spectra",
            tag=self.header,
        ):
            with dpg.group(horizontal=True, indent=8):
                self.select_all_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Select all",
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, **u
                    ),
                    user_data={},
                    tag=self.select_all_button,
                )
                self.unselect_all_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Unselect all",
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, **u
                    ),
                    user_data={},
                    tag=self.unselect_all_button,
                )
            with dpg.group(indent=8):
                self.table: int = dpg.generate_uuid()
                with dpg.table(
                    borders_outerV=True,
                    borders_outerH=True,
                    borders_innerV=True,
                    borders_innerH=True,
                    scrollY=True,
                    freeze_rows=1,
                    width=-1,
                    height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT,
                    tag=self.table,
                ):
                    dpg.add_table_column(label="", width_fixed=True)
                    attach_tooltip(tooltips.plotting.simulation_checkbox)
                    dpg.add_table_column(label="Label", width_fixed=True)
            dpg.add_spacer(height=8)
        self.sim_hash: str = ""
        self.active_hash: str = ""

    def clear(self):
        dpg.delete_item(self.table, children_only=True, slot=1)
        dpg.configure_item(self.table, height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT)
        dpg.hide_item(self.header)
        self.sim_hash = ""
        self.active_hash = ""

    def populate(
        self,
        simulations: List[SimulationResult],
        settings: PlotSettings,
    ) -> bool:
        assert type(simulations) is list, simulations
        assert type(settings) is PlotSettings, settings
        sim_hash: str = ",".join([_.uuid for _ in simulations])
        if sim_hash == self.sim_hash:
            return False
        self.clear()
        self.sim_hash = sim_hash
        if not simulations:
            dpg.configure_item(
                self.select_all_button,
                user_data={},
            )
            dpg.configure_item(
                self.unselect_all_button,
                user_data={},
            )
            return True
        sim: SimulationResult
        for sim in simulations:
            with dpg.table_row(filter_key=sim.get_label().lower(), parent=self.table):
                dpg.add_checkbox(
                    default_value=sim.uuid in settings.series_order,
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_PLOT_SERIES, enabled=a, **u
                    ),
                    user_data={
                        "simulations": [sim],
                        "settings": settings,
                    },
                )
                dpg.add_text(sim.get_label())
                attach_tooltip(sim.get_label())
        dpg.configure_item(
            self.table,
            height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT * max(1, len(simulations)),
        )
        dpg.show_item(self.header)
        return True

    def update(
        self,
        active_hash: str,
        simulations: List[SimulationResult],
        settings: PlotSettings,
    ):
        assert type(active_hash) is str, active_hash
        assert type(simulations) is list, simulations
        assert type(settings) is PlotSettings, settings
        if active_hash == self.active_hash:
            return
        self.active_hash = active_hash
        sim: SimulationResult
        row: int
        for (sim, row) in zip(simulations, dpg.get_item_children(self.table, slot=1)):
            cells: List[int] = dpg.get_item_children(row, slot=1)
            dpg.configure_item(
                cells[0],
                default_value=sim.uuid in settings.series_order,
                user_data={
                    "simulations": [sim],
                    "settings": settings,
                },
            )

    def filter(self, string: str, collapse: bool) -> List[SimulationResult]:
        assert type(string) is str, string
        stripped_string: str = string.strip()
        dpg.set_value(self.table, string)
        simulations: List[SimulationResult] = []
        row: int
        for row in dpg.get_item_children(self.table, slot=1):
            filter_key: str = dpg.get_item_filter_key(row)
            subset: List[DataSet] = dpg.get_item_user_data(
                dpg.get_item_children(row, slot=1)[0]
            ).get("simulations", [])
            if is_filtered_item_visible(row, stripped_string):
                simulations.extend(subset)
        dpg.configure_item(
            self.table,
            height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT * max(1, len(simulations)),
        )
        if simulations:
            dpg.show_item(self.header)
            if collapse:
                dpg.set_value(self.header, not string == "")
        else:
            if collapse:
                dpg.set_value(self.header, False)
            dpg.hide_item(self.header)
        dpg.get_item_user_data(self.select_all_button).update(
            {"simulations": simulations}
        )
        dpg.get_item_user_data(self.unselect_all_button).update(
            {"simulations": simulations}
        )
        dpg.set_item_label(self.header, f"Simulated spectra ({len(simulations)})")
        return simulations


class ActiveSeries:
    def __init__(self):
        self.table: int = dpg.generate_uuid()
        with dpg.table(
            borders_outerV=True,
            borders_outerH=True,
            borders_innerV=True,
            borders_innerH=True,
            scrollY=True,
            freeze_rows=1,
            width=-1,
            height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT,
            tag=self.table,
        ):
            dpg.add_table_column(
                label="Type",
                width_fixed=True,
            )
            attach_tooltip(tooltips.plotting.item_type)
            dpg.add_table_column(label="Label")
            attach_tooltip(tooltips.plotting.label)
            dpg.add_table_column(label="Appearance", width_fixed=True)
            attach_tooltip(tooltips.plotting.appearance)
            dpg.add_table_column(label="Position", width_fixed=True)
            attach_tooltip(tooltips.plotting.position)

    def clear(self):
        dpg.delete_item(self.table, children_only=True, slot=1)
        dpg.configure_item(self.table, height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT)

    def populate(
        self,
        data_sets: List[DataSet],
        tests: Dict[str, List[TestResult]],
        drts: Dict[str, List[DRTResult]],
        fits: Dict[str, List[FitResult]],
        simulations: List[SimulationResult],
        settings: PlotSettings,
    ):
        assert type(data_sets) is list, data_sets
        assert type(tests) is dict, tests
        assert type(drts) is dict, drts
        assert type(fits) is dict, fits
        assert type(simulations) is list, simulations
        assert type(settings) is PlotSettings, settings
        self.clear()
        series: List[
            Union[DataSet, TestResult, DRTResult, FitResult, SimulationResult]
        ] = []
        series.extend(filter(lambda _: _.uuid in settings.series_order, data_sets))
        series.extend(
            filter(
                lambda _: _.uuid in settings.series_order,
                list(chain(*list(tests.values()))),
            )
        )
        series.extend(
            filter(
                lambda _: _.uuid in settings.series_order,
                list(chain(*list(drts.values()))),
            )
        )
        series.extend(
            filter(
                lambda _: _.uuid in settings.series_order,
                list(chain(*list(fits.values()))),
            )
        )
        series.extend(filter(lambda _: _.uuid in settings.series_order, simulations))
        series.sort(key=lambda _: settings.series_order.index(_.uuid))
        ser: Union[DataSet, TestResult, DRTResult, FitResult, SimulationResult]
        types: dict = {
            DataSet: (
                "D",
                "Data set",
            ),
            TestResult: (
                "KK",
                "Kramers-Kronig test",
            ),
            DRTResult: (
                "DRT",
                "DRT analysis",
            ),
            FitResult: (
                "F",
                "Fit",
            ),
            SimulationResult: (
                "S",
                "Simulation",
            ),
        }
        marker_lookup: Dict[int, str] = {v: k for k, v in themes.PLOT_MARKERS.items()}
        for ser in series:
            type_label: str
            type_tooltip: str
            type_label, type_tooltip = types[type(ser)]
            with dpg.table_row(parent=self.table):
                dpg.add_text(type_label.ljust(4))
                attach_tooltip(type_tooltip)
                dpg.add_input_text(
                    hint=ser.get_label(),
                    default_value=settings.get_series_label(ser.uuid),
                    width=-1,
                    on_enter=True,
                    callback=lambda s, a, u: signals.emit(
                        Signal.RENAME_PLOT_SERIES,
                        label=a,
                        **u,
                    ),
                    user_data={
                        "settings": settings,
                        "uuid": ser.uuid,
                        "series": ser,
                    },
                )
                attach_tooltip(ser.get_label())
                dpg.add_button(
                    label="Edit",
                    width=-1,
                    callback=lambda s, a, u: signals.emit(
                        Signal.MODIFY_PLOT_SERIES_THEME,
                        settings=settings,
                        series=u,
                    ),
                    user_data=ser,
                )
                with dpg.tooltip(parent=dpg.last_item()):
                    with dpg.group(horizontal=True):
                        dpg.add_text(" Color:")
                        dpg.add_color_edit(
                            default_value=settings.get_series_color(ser.uuid),
                            enabled=False,
                            alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                            no_inputs=True,
                        )
                    dpg.add_text(
                        "Marker: "
                        + marker_lookup.get(
                            settings.get_series_marker(ser.uuid), "None"
                        )
                    )
                    dpg.add_text(
                        "  Line: "
                        + ("Yes" if settings.get_series_line(ser.uuid) else "No")
                    )
                with dpg.group(horizontal=True):
                    dpg.add_button(
                        label="U",
                        width=24,
                        callback=lambda s, a, u: signals.emit(
                            Signal.REORDER_PLOT_SERIES, settings=settings, **u
                        ),
                        user_data={
                            "uuid": ser.uuid,
                            "step": -1,
                        },
                    )
                    attach_tooltip("Move this series up by one position.")
                    dpg.add_button(
                        label="D",
                        width=24,
                        callback=lambda s, a, u: signals.emit(
                            Signal.REORDER_PLOT_SERIES, settings=settings, **u
                        ),
                        user_data={
                            "uuid": ser.uuid,
                            "step": 1,
                        },
                    )
                    attach_tooltip("Move this series down by one position.")
        dpg.configure_item(
            self.table,
            height=TABLE_HEADER_HEIGHT + TABLE_ROW_HEIGHT * max(1, len(series)),
        )

    def has_active_input(self) -> bool:
        row: int
        for row in dpg.get_item_children(self.table, slot=1):
            label_input: int = dpg.get_item_children(row, slot=1)[2]
            if dpg.is_item_active(label_input):
                return True
        return False


class PlottingTab:
    def __init__(self, state):
        self.state = state
        self.series_tags: Dict[str, int] = {}
        self.queued_update: Optional[Callable] = None
        self.plotted_uuid: str = ""
        self.plot_types: Dict[PlotType, Plot] = {}
        label_pad: int = 5
        sidebar_width: int = 400
        self.tab: int = dpg.generate_uuid()
        with dpg.tab(label="Plotting", tag=self.tab):
            with dpg.child_window(border=False):
                with dpg.group(horizontal=True):
                    self.sidebar_window: int = dpg.generate_uuid()
                    with dpg.child_window(
                        width=sidebar_width, border=False, tag=self.sidebar_window
                    ):
                        with dpg.child_window(height=82):
                            with dpg.group(horizontal=True):
                                dpg.add_text("Plot".rjust(label_pad))
                                self.plot_combo: int = dpg.generate_uuid()
                                dpg.add_combo(
                                    width=-64,
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.SELECT_PLOT_SETTINGS,
                                        settings=u.get(a),
                                    ),
                                    user_data={},
                                    tag=self.plot_combo,
                                )
                                self.new_button: int = dpg.generate_uuid()
                                dpg.add_button(
                                    label="New",
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.NEW_PLOT_SETTINGS,
                                    ),
                                    width=-1,
                                    tag=self.new_button,
                                )
                                attach_tooltip(tooltips.plotting.create)
                            with dpg.group(horizontal=True):
                                dpg.add_text("Label".rjust(label_pad))
                                self.label_input: int = dpg.generate_uuid()
                                dpg.add_input_text(
                                    width=-64,
                                    on_enter=True,
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.RENAME_PLOT_SETTINGS,
                                        label=a,
                                        settings=dpg.get_item_user_data(
                                            self.delete_button
                                        ),
                                    ),
                                    tag=self.label_input,
                                )
                                self.delete_button: int = dpg.generate_uuid()
                                dpg.add_button(
                                    label="Delete",
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.DELETE_PLOT_SETTINGS, settings=u
                                    ),
                                    user_data=None,
                                    width=-1,
                                    tag=self.delete_button,
                                )
                                attach_tooltip(tooltips.plotting.remove)
                            with dpg.group(horizontal=True):
                                dpg.add_text("Type".rjust(label_pad))
                                self.type_combo: int = dpg.generate_uuid()
                                dpg.add_combo(
                                    default_value=plot_type_to_label[PlotType.NYQUIST],
                                    items=list(
                                        map(
                                            lambda _: plot_type_to_label[_],
                                            [_ for _ in PlotType],
                                        )
                                    ),
                                    width=-64,
                                    tag=self.type_combo,
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.SELECT_PLOT_TYPE,
                                        plot_type=label_to_plot_type.get(a),
                                        settings=dpg.get_item_user_data(
                                            self.delete_button
                                        ),
                                    ),
                                    user_data=PlotType.NYQUIST,
                                )
                        with dpg.child_window(width=sidebar_width, height=-40):
                            with dpg.tab_bar():
                                with dpg.tab(label="Available"):
                                    with dpg.group(horizontal=True):
                                        dpg.add_text("Filter")
                                        attach_tooltip(tooltips.plotting.filter)
                                        self.filter_input: int = dpg.generate_uuid()
                                        dpg.add_input_text(
                                            width=-1,
                                            callback=lambda s, a, u: self.filter_possible_series(
                                                a
                                            ),
                                            tag=self.filter_input,
                                        )
                                    with dpg.child_window(
                                        border=False, width=-1, height=-1
                                    ):
                                        self.possible_series_group: int = (
                                            dpg.generate_uuid()
                                        )
                                        with dpg.group(tag=self.possible_series_group):
                                            self.possible_data_sets: DataSetsGroup = (
                                                DataSetsGroup()
                                            )
                                            self.possible_tests: TestsGroup = (
                                                TestsGroup()
                                            )
                                            self.possible_drts: DRTsGroup = DRTsGroup()
                                            self.possible_fits: FitsGroup = FitsGroup()
                                            self.possible_simulations: SimulationsGroup = (
                                                SimulationsGroup()
                                            )
                                with dpg.tab(label="Active"):
                                    self.active_series: ActiveSeries = ActiveSeries()
                        with dpg.child_window(width=sidebar_width, height=-1):
                            with dpg.group(horizontal=True):
                                self.select_all_button: int = dpg.generate_uuid()
                                dpg.add_button(
                                    label="Select all",
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.TOGGLE_PLOT_SERIES,
                                        enabled=True,
                                        **u,
                                    ),
                                    user_data={},
                                    tag=self.select_all_button,
                                )
                                attach_tooltip(tooltips.plotting.select_all)
                                self.unselect_all_button: int = dpg.generate_uuid()
                                dpg.add_button(
                                    label="Unselect all",
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.TOGGLE_PLOT_SERIES,
                                        enabled=False,
                                        **u,
                                    ),
                                    user_data={},
                                    tag=self.unselect_all_button,
                                )
                                attach_tooltip(tooltips.plotting.unselect_all)
                                self.copy_appearances_button: int = dpg.generate_uuid()
                                dpg.add_button(
                                    label="Copy appearance",
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.SELECT_PLOT_APPEARANCE_SETTINGS,
                                        settings=dpg.get_item_user_data(
                                            self.delete_button
                                        ),
                                    ),
                                    tag=self.copy_appearances_button,
                                )
                                attach_tooltip(tooltips.plotting.copy_appearance)
                                self.export_button: int = dpg.generate_uuid()
                                dpg.add_button(
                                    label="Export",
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.EXPORT_PLOT, **u
                                    ),
                                    user_data={},
                                    tag=self.export_button,
                                )
                                attach_tooltip(tooltips.plotting.export_plot)
                    with dpg.child_window(border=False, width=-1, height=-1):
                        self.plots_group: int = dpg.generate_uuid()
                        self.plot_height: int = -24
                        with dpg.group(tag=self.plots_group):
                            self.plot_types[PlotType.NYQUIST] = Nyquist(
                                width=-1,
                                height=self.plot_height,
                                legend_horizontal=False,
                                legend_location=dpg.mvPlot_Location_NorthEast,
                                legend_outside=False,
                            )
                            self.plot_types[PlotType.BODE_MAGNITUDE] = BodeMagnitude(
                                width=-1,
                                height=self.plot_height,
                                legend_horizontal=False,
                                legend_location=dpg.mvPlot_Location_NorthEast,
                                legend_outside=False,
                            )
                            self.plot_types[PlotType.BODE_PHASE] = BodePhase(
                                width=-1,
                                height=self.plot_height,
                                legend_horizontal=False,
                                legend_location=dpg.mvPlot_Location_NorthEast,
                                legend_outside=False,
                            )
                            self.plot_types[PlotType.IMPEDANCE_REAL] = ImpedanceReal(
                                width=-1,
                                height=self.plot_height,
                                legend_horizontal=False,
                                legend_location=dpg.mvPlot_Location_NorthEast,
                                legend_outside=False,
                            )
                            self.plot_types[
                                PlotType.IMPEDANCE_IMAGINARY
                            ] = ImpedanceImaginary(
                                width=-1,
                                height=self.plot_height,
                                legend_horizontal=False,
                                legend_location=dpg.mvPlot_Location_NorthEast,
                                legend_outside=False,
                            )
                            self.plot_types[PlotType.DRT] = DRT(
                                width=-1,
                                height=self.plot_height,
                                legend_horizontal=False,
                                legend_location=dpg.mvPlot_Location_NorthEast,
                                legend_outside=False,
                            )
                            self.plot_types[PlotType.BODE_MAGNITUDE].hide()
                            self.plot_types[PlotType.BODE_PHASE].hide()
                            self.plot_types[PlotType.IMPEDANCE_REAL].hide()
                            self.plot_types[PlotType.IMPEDANCE_IMAGINARY].hide()
                            self.plot_types[PlotType.DRT].hide()
                        with dpg.child_window(border=False):
                            with dpg.group(horizontal=True):
                                self.expand_sidebar_button: int = dpg.generate_uuid()
                                dpg.add_button(
                                    label="Collapse",
                                    show=True,
                                    callback=self.collapse_expand_sidebar,
                                    tag=self.expand_sidebar_button,
                                )
                                attach_tooltip(
                                    tooltips.plotting.collapse_expand_sidebar
                                )
                                self.visibility_item: int = dpg.generate_uuid()
                                self.adjust_limits_checkbox: int = dpg.generate_uuid()
                                dpg.add_checkbox(
                                    default_value=True,
                                    tag=self.adjust_limits_checkbox,
                                )
                                attach_tooltip(tooltips.general.adjust_limits)
                                dpg.add_button(
                                    label="Copy as CSV",
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.COPY_PLOT_DATA,
                                        context=Context.PLOTTING_TAB,
                                        plot=self.plot_types[
                                            dpg.get_item_user_data(self.type_combo)
                                        ],
                                    ),
                                    tag=self.visibility_item,
                                )
                                attach_tooltip(tooltips.general.copy_plot_data_as_csv)

    def resize(self, width: int, height: int):
        assert type(width) is int and width > 0
        assert type(height) is int and height > 0

    def is_visible(self) -> bool:
        return dpg.is_item_visible(self.visibility_item)

    def collapse_expand_sidebar(self):
        if self.is_sidebar_shown():
            dpg.set_item_label(self.expand_sidebar_button, "Expand")
            dpg.hide_item(self.sidebar_window)
        else:
            dpg.set_item_label(self.expand_sidebar_button, "Collapse")
            dpg.show_item(self.sidebar_window)

    def is_sidebar_shown(self) -> bool:
        return dpg.is_item_shown(self.sidebar_window)

    def populate_plots(self, plots: Dict[str, PlotSettings]):
        assert type(plots) is dict, plots
        labels: List[str] = list(plots.keys())
        label: str = dpg.get_value(self.plot_combo) or ""
        update_plot: bool = label not in labels
        if update_plot and labels:
            label = labels[0]
        dpg.configure_item(
            self.plot_combo,
            default_value=label,
            items=labels,
            user_data=plots,
        )
        if not plots:
            signals.emit(Signal.NEW_PLOT_SETTINGS)

    def plot_series(
        self,
        data_sets: List[DataSet],
        tests: Dict[str, List[TestResult]],
        drts: Dict[str, List[DRTResult]],
        fits: Dict[str, List[FitResult]],
        simulations: List[SimulationResult],
        settings: PlotSettings,
        adjust_limits: bool = True,
    ):
        assert type(data_sets) is list, data_sets
        assert type(tests) is dict, tests
        assert type(drts) is dict, drts
        assert type(fits) is dict, fits
        assert type(simulations) is list, simulations
        assert type(settings) is PlotSettings, settings
        assert type(adjust_limits) is bool, adjust_limits
        self.series_tags.clear()
        uuid: str
        label: Optional[str]
        is_simulation: bool
        is_fit: bool
        theme: Optional[int]
        from_empty: bool = False
        changed_plot_settings: bool = self.plotted_uuid != settings.uuid
        self.plotted_uuid = settings.uuid
        visible_plots: List[Plot] = []
        hidden_plots: List[Plot] = []
        plot: Plot
        for plot in self.plot_types.values():
            if plot.is_visible():
                visible_plots.append(plot)
            else:
                hidden_plots.append(plot)
        plot_type: PlotType = settings.get_type()
        plot = self.plot_types[plot_type]
        list(map(lambda _: _.clear(), visible_plots))
        list(map(lambda _: _.clear(), hidden_plots))
        if plot in hidden_plots:
            list(map(lambda _: _.hide(), visible_plots))
            plot.show()
            from_empty = True
        elif len(plot.get_series()) == 0:
            from_empty = True
        plot.clear()
        dpg.split_frame()
        plot.set_title(settings.get_label())
        for uuid in settings.series_order:
            series: Optional[
                Union[DataSet, TestResult, DRTResult, FitResult, SimulationResult]
            ]
            series = settings.find_series(
                uuid, data_sets, tests, drts, fits, simulations
            )
            if series is None:
                settings.series_order.remove(uuid)
                continue
            label = settings.get_series_label(uuid)
            show_label: bool = True
            if label == "":
                label = series.get_label()
            elif label.strip() == "":
                label = series.get_label()
                show_label = False
            is_simulation = type(series) is SimulationResult
            is_fit = (
                type(series) is FitResult
                or type(series) is TestResult
                or type(series) is DRTResult
            )
            theme = settings.get_series_theme(uuid)
            if theme < 0:
                theme = None
            real: ndarray
            imag: ndarray
            freq: ndarray
            mag: ndarray
            phase: ndarray
            tau: ndarray
            gamma: ndarray
            if plot_type == PlotType.NYQUIST:
                if settings.get_series_marker(uuid) >= 0:
                    real, imag = series.get_nyquist_data()
                    self.series_tags[series.uuid] = plot.plot(
                        real=real,
                        imaginary=imag,
                        label=label,
                        simulation=is_simulation,
                        fit=is_fit,
                        line=False,
                        theme=theme,
                        show_label=show_label,
                    )
                    if (is_simulation or is_fit) and "num_per_decade" in signature(
                        series.get_nyquist_data
                    ).parameters:
                        real, imag = series.get_nyquist_data(
                            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
                        )
                    if settings.get_series_line(uuid):
                        plot.plot(
                            real=real,
                            imaginary=imag,
                            label=label,
                            simulation=is_simulation,
                            fit=is_fit,
                            line=True,
                            theme=theme,
                            show_label=False,
                        )
                elif settings.get_series_line(uuid):
                    if (is_simulation or is_fit) and "num_per_decade" in signature(
                        series.get_nyquist_data
                    ).parameters:
                        real, imag = series.get_nyquist_data(
                            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
                        )
                    else:
                        real, imag = series.get_nyquist_data()
                    self.series_tags[series.uuid] = plot.plot(
                        real=real,
                        imaginary=imag,
                        label=label,
                        simulation=is_simulation,
                        fit=is_fit,
                        line=True,
                        theme=theme,
                        show_label=show_label,
                    )
            elif plot_type == PlotType.BODE_MAGNITUDE:
                if settings.get_series_marker(uuid) >= 0:
                    freq, mag, _ = series.get_bode_data()
                    self.series_tags[series.uuid] = plot.plot(
                        frequency=freq,
                        magnitude=mag,
                        label=label,
                        simulation=is_simulation,
                        fit=is_fit,
                        line=False,
                        theme=theme,
                        show_label=show_label,
                    )
                    if (is_simulation or is_fit) and "num_per_decade" in signature(
                        series.get_bode_data
                    ).parameters:
                        freq, mag, _ = series.get_bode_data(
                            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
                        )
                    if settings.get_series_line(uuid):
                        plot.plot(
                            frequency=freq,
                            magnitude=mag,
                            label=label,
                            simulation=is_simulation,
                            fit=is_fit,
                            line=True,
                            theme=theme,
                            show_label=False,
                        )
                elif settings.get_series_line(uuid):
                    if (is_simulation or is_fit) and "num_per_decade" in signature(
                        series.get_bode_data
                    ).parameters:
                        freq, mag, _ = series.get_bode_data(
                            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
                        )
                    else:
                        freq, mag, _ = series.get_bode_data()
                    self.series_tags[series.uuid] = plot.plot(
                        frequency=freq,
                        magnitude=mag,
                        label=label,
                        simulation=is_simulation,
                        fit=is_fit,
                        line=True,
                        theme=theme,
                        show_label=show_label,
                    )
            elif plot_type == PlotType.BODE_PHASE:
                if settings.get_series_marker(uuid) >= 0:
                    freq, _, phase = series.get_bode_data()
                    self.series_tags[series.uuid] = plot.plot(
                        frequency=freq,
                        phase=phase,
                        label=label,
                        simulation=is_simulation,
                        fit=is_fit,
                        line=False,
                        theme=theme,
                        show_label=show_label,
                    )
                    if (is_simulation or is_fit) and "num_per_decade" in signature(
                        series.get_bode_data
                    ).parameters:
                        freq, _, phase = series.get_bode_data(
                            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
                        )
                    if settings.get_series_line(uuid):
                        plot.plot(
                            frequency=freq,
                            phase=phase,
                            label=label,
                            simulation=is_simulation,
                            fit=is_fit,
                            line=True,
                            theme=theme,
                            show_label=False,
                        )
                elif settings.get_series_line(uuid):
                    if (is_simulation or is_fit) and "num_per_decade" in signature(
                        series.get_bode_data
                    ).parameters:
                        freq, _, phase = series.get_bode_data(
                            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
                        )
                    else:
                        freq, _, phase = series.get_bode_data()
                    self.series_tags[series.uuid] = plot.plot(
                        frequency=freq,
                        phase=phase,
                        label=label,
                        simulation=is_simulation,
                        fit=is_fit,
                        line=True,
                        theme=theme,
                        show_label=show_label,
                    )
            elif plot_type == PlotType.DRT:
                if type(series) is not DRTResult:
                    continue
                if series.settings.method == DRTMethod.BHT:
                    tau, gamma = series.get_drt_data()
                    self.series_tags[series.uuid] = plot.plot(
                        tau=tau,
                        gamma=gamma,
                        label=f"{label}, real" if label is not None else label,
                        line=True,
                        theme=theme,
                        show_label=show_label,
                    )
                    tau, gamma = series.get_drt_data(imaginary=True)
                    self.series_tags[f"{series.uuid}_imaginary"] = plot.plot(
                        tau=tau,
                        gamma=gamma,
                        label=f"{label}, imag." if label is not None else label,
                        line=True,
                        theme=theme,
                        show_label=show_label,
                    )
                elif (
                    series.settings.method == DRTMethod.TR_RBF
                    and series.settings.credible_intervals is True
                ):
                    tau, mean, lower, upper = series.get_drt_credible_intervals()
                    alt_color: List[float] = settings.get_series_color(uuid).copy()
                    alt_color[-1] = themes.get_plot_series_theme_color(
                        themes.drt.credible_intervals
                    )[-1]
                    self.series_tags[f"{series.uuid}_bounds"] = plot.plot(
                        tau=tau,
                        lower=lower,
                        upper=upper,
                        label=f"{label}, 3-sigma CI" if label is not None else label,
                        theme=themes.create_plot_series_theme(
                            alt_color,
                            dpg.mvPlotMarker_Circle,
                        ),
                        show_label=True,
                    )
                    self.series_tags[f"{series.uuid}_mean"] = plot.plot(
                        tau=tau,
                        gamma=mean,
                        label=f"{label}, mean" if label is not None else label,
                        line=True,
                        theme=theme,
                        show_label=show_label,
                    )
                    tau, gamma = series.get_drt_data()
                    self.series_tags[series.uuid] = plot.plot(
                        tau=tau,
                        gamma=gamma,
                        label=label,
                        line=True,
                        theme=theme,
                        show_label=show_label,
                    )
                else:
                    tau, gamma = series.get_drt_data()
                    self.series_tags[series.uuid] = plot.plot(
                        tau=tau,
                        gamma=gamma,
                        label=label,
                        line=True,
                        theme=theme,
                        show_label=show_label,
                    )
            elif plot_type == PlotType.IMPEDANCE_REAL:
                if settings.get_series_marker(uuid) >= 0:
                    freq = series.get_frequency()
                    real = series.get_impedance().real
                    self.series_tags[series.uuid] = plot.plot(
                        x=freq,
                        y=real,
                        label=label,
                        line=False,
                        theme=theme,
                        show_label=show_label,
                    )
                    if (
                        (is_simulation or is_fit)
                        and "num_per_decade"
                        in signature(series.get_frequency).parameters
                        and "num_per_decade"
                        in signature(series.get_impedance).parameters
                    ):
                        freq = series.get_frequency(
                            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
                        )
                        real = series.get_impedance(
                            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
                        ).real
                    if settings.get_series_line(uuid):
                        plot.plot(
                            x=freq,
                            y=real,
                            label=label,
                            line=True,
                            theme=theme,
                            show_label=False,
                        )
                elif settings.get_series_line(uuid):
                    if (
                        (is_simulation or is_fit)
                        and "num_per_decade"
                        in signature(series.get_frequency).parameters
                        and "num_per_decade"
                        in signature(series.get_impedance).parameters
                    ):
                        freq = series.get_frequency(
                            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
                        )
                        real = series.get_impedance(
                            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
                        ).real
                    else:
                        freq = series.get_frequency()
                        real = series.get_impedance().real
                    self.series_tags[series.uuid] = plot.plot(
                        x=freq,
                        y=real,
                        label=label,
                        line=True,
                        theme=theme,
                        show_label=False,
                    )
            elif plot_type == PlotType.IMPEDANCE_IMAGINARY:
                if settings.get_series_marker(uuid) >= 0:
                    freq = series.get_frequency()
                    imag = -series.get_impedance().imag
                    self.series_tags[series.uuid] = plot.plot(
                        x=freq,
                        y=imag,
                        label=label,
                        line=False,
                        theme=theme,
                        show_label=show_label,
                    )
                    if (
                        (is_simulation or is_fit)
                        and "num_per_decade"
                        in signature(series.get_frequency).parameters
                        and "num_per_decade"
                        in signature(series.get_impedance).parameters
                    ):
                        freq = series.get_frequency(
                            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
                        )
                        imag = -series.get_impedance(
                            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
                        ).imag
                    if settings.get_series_line(uuid):
                        plot.plot(
                            x=freq,
                            y=imag,
                            label=label,
                            line=True,
                            theme=theme,
                            show_label=False,
                        )
                elif settings.get_series_line(uuid):
                    if (
                        (is_simulation or is_fit)
                        and "num_per_decade"
                        in signature(series.get_frequency).parameters
                        and "num_per_decade"
                        in signature(series.get_impedance).parameters
                    ):
                        freq = series.get_frequency(
                            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
                        )
                        imag = -series.get_impedance(
                            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
                        ).imag
                    else:
                        freq = series.get_frequency()
                        imag = -series.get_impedance().imag
                    self.series_tags[series.uuid] = plot.plot(
                        x=freq,
                        y=imag,
                        label=label,
                        line=True,
                        theme=theme,
                        show_label=False,
                    )
        if not adjust_limits:
            return
        if dpg.get_value(self.adjust_limits_checkbox) and (
            from_empty or changed_plot_settings
        ):
            plot.queue_limits_adjustment()

    def set_series_label(self, uuid: str, label: Optional[str]):
        assert type(uuid) is str, uuid
        assert type(label) is str or label is None, label
        assert uuid in self.series_tags, self.series_tags
        for uuid in filter(lambda _: _.startswith(uuid), self.series_tags.keys()):
            if label == "":
                dpg.set_item_label(self.series_tags[uuid], "")
            elif f"{uuid}_imaginary" in self.series_tags:
                dpg.set_item_label(self.series_tags[uuid], f"{label}, real")
            elif uuid.endswith("_imaginary"):
                dpg.set_item_label(self.series_tags[uuid], f"{label}, imag.")
            elif uuid.endswith("_mean"):
                dpg.set_item_label(self.series_tags[uuid], f"{label}, mean")
            else:
                dpg.set_item_label(self.series_tags[uuid], label)

    def get_active_plot(self) -> Optional[PlotSettings]:
        return dpg.get_item_user_data(self.delete_button)

    def get_next_plot(self) -> Optional[PlotSettings]:
        settings: Optional[PlotSettings] = self.get_active_plot()
        lookup: Optional[Dict[str, PlotSettings]] = dpg.get_item_user_data(
            self.plot_combo
        )
        if settings is None or lookup is None:
            return None
        plots: List[PlotSettings] = list(lookup.values())
        if settings not in plots:
            return None
        index: int = plots.index(settings) + 1
        return plots[index % len(plots)]

    def get_previous_plot(self) -> Optional[PlotSettings]:
        settings: Optional[PlotSettings] = self.get_active_plot()
        lookup: Optional[Dict[str, PlotSettings]] = dpg.get_item_user_data(
            self.plot_combo
        )
        if settings is None or lookup is None:
            return None
        plots: List[PlotSettings] = list(lookup.values())
        if settings not in plots:
            return None
        index: int = plots.index(settings) - 1
        return plots[index % len(plots)]

    def get_next_plot_type(self) -> Optional[PlotType]:
        settings: Optional[PlotSettings] = self.get_active_plot()
        if settings is None:
            return None
        plot_types: List[PlotType] = [_ for _ in PlotType]
        index: int = plot_types.index(settings.get_type()) + 1
        return plot_types[index % len(plot_types)]

    def get_previous_plot_type(self) -> Optional[PlotType]:
        settings: Optional[PlotSettings] = self.get_active_plot()
        if settings is None:
            return None
        plot_types: List[PlotType] = [_ for _ in PlotType]
        index: int = plot_types.index(settings.get_type()) - 1
        return plot_types[index % len(plot_types)]

    def select_plot(
        self,
        settings: PlotSettings,
        data_sets: List[DataSet],
        tests: Dict[str, List[TestResult]],
        drts: Dict[str, List[DRTResult]],
        fits: Dict[str, List[FitResult]],
        simulations: List[SimulationResult],
        adjust_limits: bool,
        plot_only: bool,
    ):
        assert type(settings) is PlotSettings, settings
        assert type(data_sets) is list, data_sets
        assert type(tests) is dict, tests
        assert type(drts) is dict, drts
        assert type(fits) is dict, fits
        assert type(simulations) is list, simulations
        assert type(adjust_limits) is bool, adjust_limits
        assert type(plot_only) is bool, plot_only
        dpg.set_item_user_data(self.delete_button, settings)
        dpg.set_item_user_data(self.export_button, {"settings": settings})
        if not self.is_visible():
            self.queued_update = lambda: self.select_plot(
                settings,
                data_sets,
                tests,
                drts,
                fits,
                simulations,
                adjust_limits,
                plot_only,
            )
            return
        self.queued_update = None
        if not plot_only:
            dpg.set_value(self.plot_combo, settings.get_label())
            dpg.set_value(self.label_input, settings.get_label())
            self.populate_possible_series(
                data_sets,
                tests,
                drts,
                fits,
                simulations,
                settings,
            )
            self.active_series.populate(
                data_sets,
                tests,
                drts,
                fits,
                simulations,
                settings,
            )
        self.select_plot_type(settings.get_type())
        dpg.split_frame()
        self.plot_series(
            data_sets,
            tests,
            drts,
            fits,
            simulations,
            settings,
            adjust_limits,
        )

    def populate_possible_series(
        self,
        data_sets: List[DataSet],
        tests: Dict[str, List[TestResult]],
        drts: Dict[str, List[DRTResult]],
        fits: Dict[str, List[FitResult]],
        simulations: List[SimulationResult],
        settings: PlotSettings,
    ):
        assert type(data_sets) is list, data_sets
        assert type(tests) is dict, tests
        assert type(drts) is dict, drts
        assert type(fits) is dict, fits
        assert type(simulations) is list, simulations
        assert type(settings) is PlotSettings, settings
        self.populate_data_sets(data_sets, settings)
        self.populate_tests(tests, data_sets, settings)
        self.populate_drts(drts, data_sets, settings)
        self.populate_fits(fits, data_sets, settings)
        self.populate_simulations(simulations, settings)
        dpg.get_item_user_data(self.select_all_button).update({"settings": settings})
        dpg.get_item_user_data(self.unselect_all_button).update({"settings": settings})

    def get_filter_string(self) -> str:
        return dpg.get_value(self.filter_input).strip().lower()

    def populate_data_sets(self, data_sets: List[DataSet], settings: PlotSettings):
        assert type(data_sets) is list, data_sets
        assert type(settings) is PlotSettings, settings
        if not self.possible_data_sets.populate(data_sets, settings):
            self.possible_data_sets.update(
                ",".join(sorted(settings.series_order)) + f",{settings.uuid}",
                data_sets,
                settings,
            )
        dpg.get_item_user_data(self.possible_data_sets.select_all_button).update(
            {
                "enabled": True,
                "data_sets": data_sets,
                "settings": settings,
            }
        )
        dpg.get_item_user_data(self.possible_data_sets.unselect_all_button).update(
            {
                "enabled": False,
                "data_sets": data_sets,
                "settings": settings,
            }
        )
        data_sets = self.possible_data_sets.filter(self.get_filter_string(), False)
        dpg.get_item_user_data(self.select_all_button).update(
            {
                "data_sets": data_sets,
                "settings": settings,
            }
        )
        dpg.get_item_user_data(self.unselect_all_button).update(
            {
                "data_sets": data_sets,
                "settings": settings,
            }
        )

    def populate_tests(
        self,
        tests: Dict[str, List[TestResult]],
        data_sets: List[DataSet],
        settings: PlotSettings,
    ):
        assert type(tests) is dict, tests
        assert type(data_sets) is list, data_sets
        assert type(settings) is PlotSettings, settings
        self.possible_tests.populate(tests, data_sets, settings)
        user_data: List[TestResult] = self.possible_tests.filter(
            self.get_filter_string(), False
        )
        dpg.get_item_user_data(self.select_all_button).update(
            {
                "tests": user_data,
                "settings": settings,
            }
        )
        dpg.get_item_user_data(self.unselect_all_button).update(
            {
                "tests": user_data,
                "settings": settings,
            }
        )

    def populate_drts(
        self,
        drts: Dict[str, List[DRTResult]],
        data_sets: List[DataSet],
        settings: PlotSettings,
    ):
        assert type(drts) is dict, drts
        assert type(data_sets) is list, data_sets
        assert type(settings) is PlotSettings, settings
        self.possible_drts.populate(drts, data_sets, settings)
        user_data: List[DRTResult] = self.possible_drts.filter(
            self.get_filter_string(),
            False,
        )
        dpg.get_item_user_data(self.select_all_button).update(
            {
                "drts": user_data,
                "settings": settings,
            }
        )
        dpg.get_item_user_data(self.unselect_all_button).update(
            {
                "drts": user_data,
                "settings": settings,
            }
        )

    def populate_fits(
        self,
        fits: Dict[str, List[FitResult]],
        data_sets: List[DataSet],
        settings: PlotSettings,
    ):
        assert type(fits) is dict, fits
        assert type(data_sets) is list, data_sets
        assert type(settings) is PlotSettings, settings
        self.possible_fits.populate(fits, data_sets, settings)
        user_data: List[FitResult] = self.possible_fits.filter(
            self.get_filter_string(), False
        )
        dpg.get_item_user_data(self.select_all_button).update(
            {
                "fits": user_data,
                "settings": settings,
            }
        )
        dpg.get_item_user_data(self.unselect_all_button).update(
            {
                "fits": user_data,
                "settings": settings,
            }
        )

    def populate_simulations(
        self, simulations: List[SimulationResult], settings: PlotSettings
    ):
        assert type(simulations) is list, simulations
        assert type(settings) is PlotSettings, settings
        if not self.possible_simulations.populate(simulations, settings):
            self.possible_simulations.update(
                ",".join(sorted(settings.series_order)) + f",{settings.uuid}",
                simulations,
                settings,
            )
        dpg.get_item_user_data(self.possible_simulations.select_all_button).update(
            {
                "enabled": True,
                "simulations": simulations,
                "settings": settings,
            }
        )
        dpg.get_item_user_data(self.possible_simulations.unselect_all_button).update(
            {
                "enabled": False,
                "simulations": simulations,
                "settings": settings,
            }
        )
        simulations = self.possible_simulations.filter(self.get_filter_string(), False)
        dpg.get_item_user_data(self.select_all_button).update(
            {
                "simulations": simulations,
                "settings": settings,
            }
        )
        dpg.get_item_user_data(self.unselect_all_button).update(
            {
                "simulations": simulations,
                "settings": settings,
            }
        )

    def select_plot_type(self, plot_type: PlotType):
        assert type(plot_type) is PlotType, plot_type
        dpg.set_value(self.type_combo, plot_type_to_label.get(plot_type))
        dpg.set_item_user_data(self.type_combo, plot_type)
        for typ, plot in self.plot_types.items():
            if typ == plot_type:
                dpg.show_item(plot._plot)
            else:
                dpg.hide_item(plot._plot)

    def filter_possible_series(
        self, string: str
    ) -> Tuple[
        List[DataSet],
        List[TestResult],
        List[DRTResult],
        List[FitResult],
        List[SimulationResult],
    ]:
        string = string.lower()
        data_sets: List[DataSet] = self.possible_data_sets.filter(string, True)
        tests: List[TestResult] = self.possible_tests.filter(string, True)
        drts: List[DRTResult] = self.possible_drts.filter(string, True)
        fits: List[FitResult] = self.possible_fits.filter(string, True)
        simulations: List[SimulationResult] = self.possible_simulations.filter(
            string, True
        )
        dpg.get_item_user_data(self.select_all_button).update(
            {
                "data_sets": data_sets,
                "tests": tests,
                "drts": drts,
                "fits": fits,
                "simulations": simulations,
            }
        )
        dpg.get_item_user_data(self.unselect_all_button).update(
            {
                "data_sets": data_sets,
                "tests": tests,
                "drts": drts,
                "fits": fits,
                "simulations": simulations,
            }
        )
        return (
            data_sets,
            tests,
            drts,
            fits,
            simulations,
        )

    def has_active_input(self) -> bool:
        return (
            dpg.is_item_active(self.label_input)
            or dpg.is_item_active(self.filter_input)
            or self.active_series.has_active_input()
        )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from uuid import uuid4 as _uuid4
from time import time as _time
import pyimpspec as _pyimpspec
from pyimpspec import (
    DRTError,
)
from pyimpspec.analysis.drt.bht import _get_default_num_procs
from deareis.data import DataSet
from deareis.data.drt import (
    DRTResult,
    DRTSettings,
)
from deareis.enums import (
    DRTMethod,
    DRTMode,
    RBFShape,
    RBFType,
    drt_method_to_value as _drt_method_to_value,
    drt_mode_to_value as _drt_mode_to_value,
    rbf_shape_to_value as _rbf_shape_to_value,
    rbf_type_to_value as _rbf_type_to_value,
)


def calculate_drt(
    data: DataSet,
    settings: DRTSettings,
    num_procs: int = -1,
) -> DRTResult:
    """
    Wrapper for the `pyimpspec.calculate_drt` function.

    Calculates the distribution of relaxation times (DRT) for a given data set.

    References:

    - Kulikovsky, A., 2020, Phys. Chem. Chem. Phys., 22, 19131-19138 (https://doi.org/10.1039/D0CP02094J)
    - Wan, T. H., Saccoccio, M., Chen, C., and Ciucci, F., 2015, Electrochim. Acta, 184, 483-499 (https://doi.org/10.1016/j.electacta.2015.09.097).
    - Ciucci, F. and Chen, C., 2015, Electrochim. Acta, 167, 439-454 (https://doi.org/10.1016/j.electacta.2015.03.123)
    - Effat, M. B. and Ciucci, F., 2017, Electrochim. Acta, 247, 1117-1129 (https://doi.org/10.1016/j.electacta.2017.07.050)
    - Liu, J., Wan, T. H., and Ciucci, F., 2020, Electrochim. Acta, 357, 136864 (https://doi.org/10.1016/j.electacta.2020.136864)
    - Boukamp, B.A., 2015, Electrochim. Acta, 154, 35-46, (https://doi.org/10.1016/j.electacta.2014.12.059)
    - Boukamp, B.A. and Rolle, A, 2017, Solid State Ionics, 302, 12-18 (https://doi.org/10.1016/j.ssi.2016.10.009)


    Parameters
    ----------
    data: DataSet
        The data set to use in the calculations.

    settings: DRTSettings
        The settings to use.

    num_procs: int = -1
        The maximum number of processes to use.
        A value below one results in using the total number of CPU cores present.
    """
    if settings.method == DRTMethod.M_RQ_FIT:
        assert settings.circuit is not None, "A (fitted) circuit has not been provided!"
    if num_procs < 1:
        num_procs = _get_default_num_procs()
    result: _pyimpspec.DRTResult = _pyimpspec.calculate_drt(
        data=data,
        method=_drt_method_to_value[settings.method],
        mode=_drt_mode_to_value[settings.mode],
        lambda_value=settings.lambda_value,
        rbf_type=_rbf_type_to_value[settings.rbf_type],
        derivative_order=settings.derivative_order,
        rbf_shape=_rbf_shape_to_value[settings.rbf_shape],
        shape_coeff=settings.shape_coeff,
        inductance=settings.inductance,
        credible_intervals=settings.credible_intervals,
        num_samples=settings.num_samples,
        num_attempts=settings.num_attempts,
        maximum_symmetry=settings.maximum_symmetry,
        circuit=settings.circuit,
        W=settings.W,
        num_per_decade=settings.num_per_decade,
        num_procs=num_procs,
    )
    return DRTResult(
        _uuid4().hex,
        _time(),
        result.tau,
        result.gamma,
        result.frequency,
        result.impedance,
        result.real_residual,
        result.imaginary_residual,
        result.mean_gamma,
        result.lower_bound,
        result.upper_bound,
        result.imaginary_gamma,
        result.scores,
        result.chisqr,
        result.lambda_value,
        data.get_mask().copy(),
        settings,
    )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from os import (
    getcwd,
    makedirs,
    walk,
)
from os.path import (
    basename,
    exists,
    getmtime,
    getsize,
    islink,
    join,
    split,
    splitext,
)
from pathlib import Path
from typing import (
    Callable,
    Dict,
    List,
    Optional,
)
import dearpygui.dearpygui as dpg
from deareis.tooltips import attach_tooltip
from deareis.utility import (
    calculate_window_position_dimensions,
    format_timestamp,
    is_filtered_item_visible,
)
from deareis.keybindings import is_control_down, is_shift_down
from deareis.signals import Signal
import deareis.signals as signals
import deareis.themes as themes


class FileDialog:
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        cwd: str = kwargs.get("cwd", getcwd())
        assert type(cwd) is str, cwd
        if not exists(cwd):
            cwd = getcwd()
        self._cwd: str = cwd
        label: str = kwargs.get("label", "File dialog")
        assert type(label) is str, label
        extensions: List[str] = kwargs.get("extensions", [".*"])
        assert type(extensions) is list and len(extensions) > 0, extensions
        default_extension: str = kwargs.get("default_extension", extensions[0])
        assert default_extension in extensions, (
            default_extension,
            extensions,
        )
        self._callback: Callable = kwargs["callback"]
        self._save: bool = kwargs.get("save", False)
        self._merge: bool = kwargs.get("merge", False)
        self._window: int = dpg.generate_uuid()
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions()
        with dpg.window(
            label=label,
            modal=True,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
            show=False,
            on_close=self.close,
            tag=self._window,
        ):
            with dpg.group(horizontal=True):
                if self._save:
                    dpg.add_button(label="N", callback=lambda: self.create_directory())
                    attach_tooltip("Create a new directory.")
                dpg.add_button(
                    label="R",
                    callback=lambda: self.reset_path(),
                )
                attach_tooltip(f"Reset to current working directory: '{self._cwd}'.")
                dpg.add_button(label="E", callback=lambda: self.edit_path())
                attach_tooltip("Edit the path via input.")
                self._path_combo: int = dpg.generate_uuid()
                dpg.add_combo(
                    tag=self._path_combo,
                    width=-1,
                    callback=lambda s, a, u: self.update_current_path(
                        u.get(a, self._cwd)
                    ),
                )
                self._path_input: int = dpg.generate_uuid()
                dpg.add_input_text(
                    hint="Path...",
                    callback=lambda s, a, u: self.update_current_path(a),
                    on_enter=True,
                    show=False,
                    width=-1,
                    tag=self._path_input,
                )
            with dpg.group(horizontal=True):
                dpg.add_button(label="C", callback=lambda: self.clear_search())
                attach_tooltip("Clear the search input.")
                self._search_input: int = dpg.generate_uuid()
                dpg.add_input_text(
                    hint="Search...",
                    width=-100 if not self._save else -1,
                    tag=self._search_input,
                    callback=lambda s, a, u: dpg.set_value(self._table, a.lower()),
                )
                self._extension_combo: int = dpg.generate_uuid()
                dpg.add_combo(
                    default_value=default_extension,
                    items=extensions,
                    callback=lambda: self.update_current_path(self.get_current_path()),
                    show=not self._save,
                    tag=self._extension_combo,
                    width=-1,
                )
            self._table: int = dpg.generate_uuid()
            with dpg.table(
                borders_outerV=True,
                borders_outerH=True,
                borders_innerV=True,
                borders_innerH=True,
                scrollY=True,
                freeze_rows=1,
                height=-24,
                tag=self._table,
            ):
                if not self._save:
                    dpg.add_table_column(
                        label="",
                        width_fixed=True,
                        width=24,
                    )
                dpg.add_table_column(
                    label="",
                    width_fixed=True,
                    width=24,
                )
                dpg.add_table_column(
                    label="Name",
                )
                dpg.add_table_column(
                    label="Size",
                    width_fixed=True,
                    width=200,
                )
                dpg.add_table_column(
                    label="Modified",
                    width_fixed=True,
                    width=200,
                )
            with dpg.group(horizontal=True):
                if not self._save:
                    dpg.add_button(
                        label=("Merge" if self._merge else "Load"),
                        callback=lambda: self.load_files(),
                    )
                    dpg.add_button(
                        label="Select all",
                        callback=lambda: self.select_files(state=True),
                    )
                    dpg.add_button(
                        label="Unselect all",
                        callback=lambda: self.select_files(state=False),
                    )
                else:
                    dpg.add_button(label="Save", callback=lambda: self.save_file())
                    self._name_input: int = dpg.generate_uuid()
                    dpg.add_input_text(
                        hint="Name...",
                        callback=lambda: self.save_file(),
                        on_enter=True,
                        width=-100,
                        tag=self._name_input,
                    )
                    self._name_extension_combo: int = dpg.generate_uuid()
                    dpg.add_combo(
                        default_value=default_extension,
                        items=extensions,
                        width=-1,
                        callback=lambda: self.update_current_path(
                            self.get_current_path()
                        ),
                        tag=self._name_extension_combo,
                    )
        self.update_current_path(self._cwd)
        self.show()

    def hide(self):
        dpg.hide_item(self._window)
        if dpg.does_item_exist(self._key_handler):
            dpg.delete_item(self._key_handler)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)

    def show(self):
        self._key_handler: int = dpg.generate_uuid()
        with dpg.handler_registry(tag=self._key_handler):
            dpg.add_key_release_handler(
                key=dpg.mvKey_Escape,
                callback=self.close,
            )
            if self._save:
                dpg.add_key_release_handler(
                    key=dpg.mvKey_N,
                    callback=lambda: self.create_directory(keybinding=True),
                )
            else:
                dpg.add_key_release_handler(
                    key=dpg.mvKey_A,
                    callback=lambda: self.select_files(keybinding=True),
                )
            dpg.add_key_release_handler(
                key=dpg.mvKey_R,
                callback=lambda: self.reset_path(keybinding=True),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_E,
                callback=lambda: self.edit_path(keybinding=True),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_F,
                callback=lambda: self.focus_search(keybinding=True),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_C,
                callback=lambda: self.clear_search(keybinding=True),
            )
        dpg.show_item(self._window)
        if self._save:
            dpg.focus_item(self._name_input)
        signals.emit(Signal.BLOCK_KEYBINDINGS, window=self._window, window_object=None)

    def close(self):
        self.hide()
        dpg.delete_item(self._window)

    def create_directory(self, keybinding: bool = False):
        assert type(keybinding) is bool, keybinding
        if keybinding and not is_control_down():
            return
        self.hide()
        dpg.split_frame(delay=33)
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions(400, 50)
        key_handler: int = dpg.generate_uuid()
        window: int = dpg.generate_uuid()
        name_input: int = dpg.generate_uuid()

        def close(path: str = ""):
            dpg.delete_item(window)
            dpg.delete_item(key_handler)
            dpg.split_frame()
            self.show()
            if path != "":
                self.update_current_path(path)

        def accept():
            name: str = dpg.get_value(name_input).strip()
            path: str = join(self.get_current_path(), name)
            if name != "" and not exists(path):
                try:
                    makedirs(path)
                except Exception:
                    pass
                close(path)
            else:
                close()

        with dpg.handler_registry(tag=key_handler):
            dpg.add_key_release_handler(
                key=dpg.mvKey_Escape,
                callback=close,
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Return,
                callback=accept,
            )

        with dpg.window(
            label="Create folder",
            modal=True,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
            show=False,
            on_close=close,
            tag=window,
        ):
            dpg.add_input_text(hint="Name...", width=-1, tag=name_input)
            dpg.add_button(label="Accept", callback=accept)

        dpg.show_item(window)
        dpg.split_frame()
        dpg.focus_item(name_input)
        signals.emit(Signal.BLOCK_KEYBINDINGS, window=window, window_object=None)

    def clear_search(self, keybinding: bool = False):
        assert type(keybinding) is bool, keybinding
        if keybinding and not is_control_down():
            return
        dpg.set_value(self._table, "")
        dpg.set_value(self._search_input, "")

    def focus_search(self, keybinding: bool = False):
        if keybinding and not is_control_down():
            return
        dpg.focus_item(self._search_input)

    def select_files(self, state: Optional[bool] = None, keybinding: bool = False):
        assert type(state) is bool or state is None, state
        assert type(keybinding) is bool, keybinding
        if keybinding:
            if not is_control_down():
                return
            if dpg.is_item_focused(self._path_input) or dpg.is_item_focused(
                self._search_input
            ):
                return
            state = not is_shift_down()
        assert state is not None
        filter_key: str = dpg.get_value(self._search_input).lower()
        files: Dict[int, Optional[str]] = {}
        row: int
        for row in dpg.get_item_children(self._table, slot=1):
            if not is_filtered_item_visible(row, filter_key):
                continue
            size_text: int = dpg.get_item_children(row, slot=1)[-2]
            if dpg.get_value(size_text) == "":
                continue
            files[row] = dpg.get_item_user_data(row)
        path: Optional[str]
        for row, path in files.items():
            checkbox: int = dpg.get_item_children(row, slot=1)[0]
            assert "mvCheckbox" in dpg.get_item_type(checkbox), dpg.get_item_type(
                checkbox
            )
            if state and path is None:
                dpg.set_value(checkbox, state)
                dpg.set_item_user_data(row, dpg.get_item_user_data(checkbox)[1])
            elif not state and path is not None:
                dpg.set_value(checkbox, state)
                dpg.set_item_user_data(row, None)

    def get_current_path(self) -> str:
        return dpg.get_value(self._path_combo)

    def reset_path(self, keybinding: bool = False):
        assert type(keybinding) is bool, keybinding
        if keybinding and not is_control_down():
            return
        self.update_current_path(self._cwd)

    def edit_path(self, keybinding: bool = False):
        assert type(keybinding) is bool, keybinding
        if keybinding and not is_control_down():
            return
        dpg.hide_item(self._path_combo)
        dpg.show_item(self._path_input)
        dpg.set_value(self._path_input, self.get_current_path())
        dpg.focus_item(self._path_input)

    def update_current_path(self, path: str):
        if not exists(path):
            return
        self.clear_search()
        self.update_path_combo(path)
        self.update_contents_table(path)

    def update_path_combo(self, path: str):
        dpg.hide_item(self._path_input)
        lookup: Dict[str, str] = {}
        items: List[str] = []
        a: str
        b: str
        suffix: str = ""
        a, b = split(path)
        while b != "":
            items.insert(0, b + suffix)
            lookup[b + suffix] = join(a, b)
            a, b = split(a)
            suffix += " "
        items.insert(0, a + suffix)
        lookup[a + suffix] = a
        dpg.configure_item(
            self._path_combo,
            default_value=path,
            items=items,
            user_data=lookup,
        )
        dpg.show_item(self._path_combo)

    def update_contents_table(self, root: str):
        extension_filter: str = dpg.get_value(
            self._extension_combo if not self._save else self._name_extension_combo
        ).lower()
        dpg.delete_item(self._table, children_only=True, slot=1)
        for _, dirs, files in walk(root):
            break
        dirs.sort()
        files.sort()
        dpg.push_container_stack(self._table)
        path: str
        directory: str
        for directory in dirs:
            path = join(root, directory)
            with dpg.table_row(filter_key=directory.lower()):
                if not self._save:
                    dpg.add_checkbox(enabled=False, show=False)
                if islink(path):
                    path = str(Path(path).resolve())
                    dpg.add_text("L")
                    attach_tooltip(f"Link to directory: '{path}'")
                else:
                    dpg.add_text("D")
                    attach_tooltip("Directory")
                dpg.bind_item_theme(
                    dpg.add_button(
                        label=directory,
                        callback=lambda s, a, u: self.update_current_path(u),
                        user_data=path,
                        width=-1,
                    ),
                    themes.file_dialog.folder_button,
                )
                dpg.add_text("", show=False)
                dpg.add_text("", show=False)
        for file in files:
            if not (
                extension_filter == ".*" or file.lower().endswith(extension_filter)
            ):
                continue
            path = join(root, file)
            link: bool = False
            if islink(path):
                path = str(Path(path).resolve())
                if not exists(path):
                    continue
                link = True
            row: int
            with dpg.table_row(filter_key=file.lower()) as row:
                if not self._save:
                    dpg.add_checkbox(
                        callback=lambda s, a, u: dpg.set_item_user_data(
                            u[0], u[1] if a else None
                        ),
                        user_data=(
                            row,
                            path,
                        ),
                    )
                    attach_tooltip(
                        "Select multiple files to "
                        + ("merge." if self._merge else "load.")
                    )
                if link:
                    dpg.add_text("L")
                    attach_tooltip(f"Link to file: '{path}'")
                else:
                    dpg.add_text("F")
                    attach_tooltip("File")
                dpg.bind_item_theme(
                    dpg.add_button(
                        label=file,
                        callback=lambda s, a, u: self.click_file(u),
                        user_data=path,
                        width=-1,
                    ),
                    themes.file_dialog.file_button,
                )
                dpg.add_text(self.format_size(getsize(path)))
                dpg.add_text(format_timestamp(getmtime(path)))
        dpg.pop_container_stack()

    def format_size(self, num_bytes: int) -> str:
        suffixes: List[str] = [
            "B",
            "KiB",
            "MiB",
            "GiB",
            "TiB",
            "PiB",
            "EiB",
            "ZiB",
            "YiB",
        ]
        i: int = 0
        value: float = float(num_bytes)
        while value > 1.0:
            i += 1
            value /= 1024
        i -= 1
        if i < 0:
            i = 0
        return f"{float(num_bytes) / pow(1024, i):.3g} {suffixes[i]}".ljust(9)

    def click_file(self, path: str):
        if not self._save:
            if self._merge:
                return
            self.hide()
            self._callback(paths=[path])
            self.close()
        else:
            dpg.set_value(self._name_input, splitext(basename(path))[0])

    def load_files(self):
        paths: List[str] = list(
            filter(
                lambda _: _ is not None,
                map(
                    lambda _: dpg.get_item_user_data(_),
                    dpg.get_item_children(self._table, slot=1),
                ),
            )
        )
        if not paths:
            return
        self.hide()
        self._callback(paths=paths, merge=self._merge)
        self.close()

    def save_file(self):
        name: str = dpg.get_value(self._name_input).strip()
        if name == "":
            dpg.focus_item(self._name_input)
            return
        self.hide()
        path: str = join(self.get_current_path(), name)
        extension: str = dpg.get_value(self._name_extension_combo)
        if not path.endswith(extension):
            path += extension
        self._callback(path=path)
        self.close()



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from types import SimpleNamespace


data_sets = SimpleNamespace(
    **{
        "load": """
Select files to load as new data sets.
    """.strip(),
        "delete": """
Delete the current data set.
    """.strip(),
        "average": """
Create a new data set by averaging multiple data sets.
    """.strip(),
        "toggle": """
Toggle a range of points.
    """.strip(),
        "copy": """
Copy the mask from another data set.
    """.strip(),
        "copy_mask_source": """
Select the data set to copy the mask from. The 'Page up' and 'Page down' keys can also be used for to cycle between data sets.
    """.strip(),
        "subtract": """
Subtract impedance from the current spectrum.
    """.strip(),
        "mask": """
If checked, then the corresponding data point will be omitted from plots, tests, and fits.
    """.strip(),
        "frequency": """
The excitation frequency that was applied when recording the data point.
    """.strip(),
        "real": """
The real part of the complex impedance.
    """.strip(),
        "imaginary": """
The negative imaginary part of the complex impedance.
    """.strip(),
        "magnitude": """
The magnitude of the complex impedance.
    """.strip(),
        "phase": """
The negative phase shift of the complex impedance.
    """.strip(),
    }
)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

import dearpygui.dearpygui as dpg

dpg.create_context()

import matplotlib

matplotlib.use("Agg")

from json import (
    dumps as dump_json,
    load as load_json,
)
from multiprocessing import (
    Process,
    Queue,
    set_start_method,
)
from queue import Empty
from os import remove
from traceback import format_exc
from typing import (
    Dict,
    IO,
    List,
    Optional,
    Union,
)
from numpy import (
    array,
    ndarray,
)
import webbrowser
from pyimpspec import (
    Circuit,
)
import pyimpspec
from pandas import DataFrame
from sympy import (
    Expr,
    latex,
    simplify,
)
from .project import (
    close_project,
    create_project_snapshot,
    load_project_files,
    new_project,
    restore_project_state,
    save_project,
    save_project_as,
    select_project_files,
)
from .overview import (
    rename_project,
    modify_project_notes,
)
from .data_sets import (
    apply_data_set_mask,
    delete_data_set,
    load_data_set_files,
    load_simulation_as_data_set,
    modify_data_set_path,
    rename_data_set,
    select_data_points_to_toggle,
    select_data_set,
    select_data_set_files,
    select_data_set_mask_to_copy,
    select_data_sets_to_average,
    select_impedance_to_subtract,
    toggle_data_point,
)
from .kramers_kronig import (
    apply_test_settings,
    delete_test_result,
    perform_test,
    select_test_result,
)
from .drt import (
    apply_drt_settings,
    delete_drt_result,
    perform_drt,
    select_drt_result,
)
from .fitting import (
    apply_fit_settings,
    delete_fit_result,
    perform_fit,
    select_fit_result,
)
from .simulation import (
    apply_simulation_settings,
    delete_simulation_result,
    perform_simulation,
    select_simulation_result,
)
from .plotting import (
    copy_plot_appearance_settings,
    delete_plot_settings,
    export_plot,
    modify_plot_series_theme,
    new_plot_settings,
    rename_plot_series,
    rename_plot_settings,
    reorder_plot_series,
    save_plot,
    select_plot_appearance_settings,
    select_plot_settings,
    select_plot_type,
    toggle_plot_series,
)
from .check_updates import perform_update_check
from deareis.gui.plots import show_modal_plot_window
from deareis.gui.changelog import show_changelog
from deareis.enums import (
    Context,
    DRTOutput,
    FitSimOutput,
    drt_output_to_label,
    label_to_drt_output,
    fit_sim_output_to_label,
)
from deareis.data import (
    DRTResult,
    DataSet,
    FitResult,
    Project,
    SimulationResult,
)
from deareis.arguments import (
    Namespace,
    parse as parse_arguments,
)
from deareis.gui import (
    ProjectTab,
)
from deareis.signals import Signal
import deareis.signals as signals
from deareis.state import STATE
from deareis.utility import (
    calculate_window_position_dimensions,
    pad_dataframe_dictionary,
)
from deareis.gui.plots import (
    Bode,
    BodeMagnitude,
    BodePhase,
    DRT,
    Impedance,
    ImpedanceReal,
    ImpedanceImaginary,
    Nyquist,
    Residuals,
)
from deareis.gui.plots.base import Plot
from deareis.gui.licenses import show_license_window
from deareis.gui.settings import (
    AppearanceSettings,
    KeybindingRemapping,
    show_defaults_settings_window,
)
import deareis.themes as themes
from deareis.version import PACKAGE_VERSION


# Hook into the progress callbacks implemented in pyimpspec
pyimpspec.progress.register(
    lambda *a, **k: signals.emit(Signal.SHOW_BUSY_MESSAGE, *a, **k)
)


def sympy_wrapper(expr: Expr, queue: Queue):
    queue.put(simplify(expr))


def get_sympy_expr(circuit: Circuit) -> Expr:
    assert type(circuit) is Circuit
    expr: Expr = circuit.to_sympy()
    # Try to simplify the expression, but don't wait for an indefinite period of time
    queue: Queue = Queue()
    proc: Process = Process(
        target=sympy_wrapper,
        args=(
            expr,
            queue,
        ),
    )
    proc.start()
    try:
        expr = queue.get(True, 2)
    except Empty:
        pass
    if proc.is_alive:
        proc.kill()
    return expr


def copy_output(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    output: Optional[Union[FitSimOutput, DRTOutput]] = kwargs.get("output")
    if output is None:
        return
    clipboard_content: str = ""
    if type(output) is FitSimOutput:
        fit_or_sim: Optional[Union[FitResult, SimulationResult]] = kwargs.get(
            "fit_or_sim"
        )
        data: Optional[DataSet] = kwargs.get("data")
        if fit_or_sim is None:
            return
        assert output in fit_sim_output_to_label, "Unsupported output!"
        signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Generating output")
        if output == FitSimOutput.CDC_BASIC:
            clipboard_content = fit_or_sim.circuit.to_string()
        elif output == FitSimOutput.CDC_EXTENDED:
            clipboard_content = fit_or_sim.circuit.to_string(6)
        elif output == FitSimOutput.CSV_DATA_TABLE:
            Z_fit_or_sim: ndarray = fit_or_sim.get_impedance()
            dictionary: dict = {}
            if type(fit_or_sim) is FitResult:
                Z_exp: ndarray = data.get_impedance(masked=None)
                indices: ndarray = array(
                    [
                        _
                        for _ in range(0, data.get_num_points(masked=None))
                        if fit_or_sim.mask.get(_, False) is False
                    ]
                )
                dictionary = {
                    "f (Hz)": fit_or_sim.get_frequency(),
                    "Zre_exp (ohm)": Z_exp[indices].real,
                    "Zim_exp (ohm)": Z_exp[indices].imag,
                    "Zre_fit (ohm)": Z_fit_or_sim.real,
                    "Zim_fit (ohm)": Z_fit_or_sim.imag,
                }
            else:
                dictionary = {
                    "f (Hz)": fit_or_sim.get_frequency(),
                    "Zre_sim (ohm)": Z_fit_or_sim.real,
                    "Zim_sim (ohm)": Z_fit_or_sim.imag,
                }
            if dictionary:
                dataframe = DataFrame.from_dict(dictionary)
                clipboard_content = dataframe.to_csv(index=False)
        elif (
            output == FitSimOutput.CSV_PARAMETERS_TABLE
            or output == FitSimOutput.JSON_PARAMETERS_TABLE
            or output == FitSimOutput.LATEX_PARAMETERS_TABLE
            or output == FitSimOutput.MARKDOWN_PARAMETERS_TABLE
        ):
            dataframe = fit_or_sim.to_dataframe()
            if output == FitSimOutput.CSV_PARAMETERS_TABLE:
                clipboard_content = dataframe.to_csv(index=False)
            elif output == FitSimOutput.JSON_PARAMETERS_TABLE:
                clipboard_content = dataframe.to_json()
            elif output == FitSimOutput.LATEX_PARAMETERS_TABLE:
                clipboard_content = (
                    dataframe.style.format(
                        {
                            "Value": "{:.3g}",
                            "Std. err. (%)": "{:.3g}",
                        }
                    )
                    .format_index(axis="columns", escape="latex")
                    .hide(axis="index")
                    .to_latex(hrules=True)
                )
            elif output == FitSimOutput.MARKDOWN_PARAMETERS_TABLE:
                clipboard_content = dataframe.to_markdown(
                    index=False,
                    floatfmt=".3g",
                )
        elif output == FitSimOutput.LATEX_DIAGRAM:
            clipboard_content = fit_or_sim.circuit.to_circuitikz()
        elif output == FitSimOutput.SVG_DIAGRAM:
            clipboard_content = (
                fit_or_sim.circuit.to_drawing().get_imagedata(fmt="svg").decode()
            )
        elif output == FitSimOutput.SVG_DIAGRAM_NO_TERMINAL_LABELS:
            clipboard_content = (
                fit_or_sim.circuit.to_drawing(working_label="", counter_label="")
                .get_imagedata(fmt="svg")
                .decode()
            )
        elif output == FitSimOutput.SVG_DIAGRAM_NO_LABELS:
            clipboard_content = (
                fit_or_sim.circuit.to_drawing(hide_labels=True)
                .get_imagedata(fmt="svg")
                .decode()
            )
        elif output == FitSimOutput.LATEX_EXPR:
            clipboard_content = latex(get_sympy_expr(fit_or_sim.circuit))
        elif (
            output == FitSimOutput.SYMPY_EXPR
            or output == FitSimOutput.SYMPY_EXPR_VALUES
        ):
            expr = get_sympy_expr(fit_or_sim.circuit)
            if output == FitSimOutput.SYMPY_EXPR:
                clipboard_content = str(expr)
            else:
                lines: List[str] = []
                lines.append(f'expr = sympify("{str(expr)}")')
                symbols: List[str] = list(sorted(map(str, expr.free_symbols)))
                if len(symbols) == 0:
                    clipboard_content = str(expr)
                else:
                    parameters = fit_or_sim.circuit.get_parameters()
                    lines.append(
                        ", ".join(symbols) + " = sorted(expr.free_symbols, key=str)"
                    )
                    lines.append("parameters = {")
                    if "f" in symbols:
                        symbols.remove("f")
                    sym: str
                    for sym in symbols:
                        assert "_" in sym
                        ident: Union[int, str]
                        label, ident = sym.split("_")
                        value: Optional[float] = None
                        try:
                            ident = int(label)
                            assert ident in parameters
                            value = parameters[ident][label]
                        except ValueError:
                            for element in fit_or_sim.circuit.get_elements():
                                if not element.get_label().endswith(f"_{ident}"):
                                    continue
                                value = element.get_parameters().get(label)
                        assert value is not None
                        lines.append(f"\t{sym}: {value:.6E},")
                    lines.append("}")
                    clipboard_content = "\n".join(lines)
        else:
            raise Exception(f"Unsupported FitSimOutput: {output=}")
    elif type(output) is DRTOutput:
        assert output in drt_output_to_label, "Unsupported output!"
        drt: Optional[DRTResult] = kwargs.get("drt")
        if drt is None:
            return
        signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Generating output")
        if (
            output == DRTOutput.CSV_SCORES
            or output == DRTOutput.JSON_SCORES
            or output == DRTOutput.LATEX_SCORES
            or output == DRTOutput.MARKDOWN_SCORES
        ):
            score_dataframe: Optional[DataFrame] = drt.get_score_dataframe(
                latex_labels=output == DRTOutput.LATEX_SCORES
            )
            if score_dataframe is not None:
                if output == DRTOutput.CSV_SCORES:
                    clipboard_content = score_dataframe.to_csv(index=False)
                elif output == DRTOutput.JSON_SCORES:
                    clipboard_content = score_dataframe.to_json()
                elif output == DRTOutput.LATEX_SCORES:
                    clipboard_content = (
                        score_dataframe.style.format(
                            {
                                "Real (\%)": "{:.3g}",
                                "Imaginary (\%)": "{:.3g}",
                            }
                        )
                        .hide(axis="index")
                        .to_latex(hrules=True)
                    )
                elif output == DRTOutput.MARKDOWN_SCORES:
                    clipboard_content = score_dataframe.to_markdown(
                        index=False,
                        floatfmt=".3g",
                    )
    else:
        raise Exception(f"Unsupported output type: {type(output)}")
    if clipboard_content != "":
        dpg.set_clipboard_text(clipboard_content)
    signals.emit(Signal.HIDE_BUSY_MESSAGE)


def undo(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Undoing")
    signals.emit(
        Signal.RESTORE_PROJECT_STATE,
        project=project,
        project_tab=project_tab,
        state_snapshot=STATE.get_previous_project_state_snapshot(project),
    )
    signals.emit(Signal.HIDE_BUSY_MESSAGE)


def redo(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Redoing")
    signals.emit(
        Signal.RESTORE_PROJECT_STATE,
        project=project,
        project_tab=project_tab,
        state_snapshot=STATE.get_next_project_state_snapshot(project),
    )
    signals.emit(Signal.HIDE_BUSY_MESSAGE)


def select_home_tab(*args, **kwargs):
    STATE.set_active_project(None)


def select_project_tab(*args, **kwargs):
    uuid: Optional[str] = kwargs.get("uuid")
    assert type(uuid) is str or uuid is None
    STATE.set_active_project(uuid)
    signals.emit(
        Signal.VIEWPORT_RESIZED,
        width=dpg.get_viewport_width(),
        height=dpg.get_viewport_height(),
    )


def viewport_resized(width: int, height: int):
    STATE.program_window.busy_message.resize(width, height)
    STATE.program_window.error_message.resize(width, height)
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project_tab is not None:
        project_tab.resize(width, height)
    modal_window: Optional[int] = STATE.get_active_modal_window()
    if modal_window is None or not dpg.does_item_exist(modal_window):
        return
    elif modal_window in [
        STATE.program_window.busy_message.window,
        STATE.program_window.error_message.window,
    ]:
        return
    item_configuration: dict = dpg.get_item_configuration(modal_window)
    if item_configuration.get("no_move", False):
        return
    x: int
    y: int
    w: int
    h: int
    if item_configuration.get("no_resize", False):
        x, y, w, h = calculate_window_position_dimensions(
            dpg.get_item_width(modal_window), dpg.get_item_height(modal_window)
        )
    else:
        x, y, w, h = calculate_window_position_dimensions()
    if "mvFileDialog" in dpg.get_item_type(modal_window):
        dpg.configure_item(
            modal_window,
            width=w,
            height=h,
        )
    else:
        dpg.configure_item(
            modal_window,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
        )


def show_enlarged_plot(*args, **kwargs):
    plot: Optional[Plot] = kwargs.get("plot")
    adjust_limits: bool = kwargs.get("adjust_limits", True)
    if plot is None or plot.is_blank():
        return
    show_modal_plot_window(plot, adjust_limits=adjust_limits)


def copy_plot_data(*args, **kwargs):
    plot: Optional[Plot] = kwargs.get("plot")
    context: Optional[Context] = kwargs.get("context")
    if plot is None or plot.is_blank() or context is None:
        return
    dictionary: dict = {}
    series: dict
    label: str
    key: str
    if type(plot) is Bode:
        for series in plot.get_series():
            label = "Data"
            if series.get("simulation", False):
                label = "Sim."
            elif series.get("fit", False):
                label = "Fit"
            if series.get("line", False):
                label += " (line)"
            else:
                label += " (scatter)"
            key = f"f (Hz) - {label}"
            while key in dictionary:
                label += " "
                key = f"f (Hz) - {label}"
            dictionary[key] = series["frequency"]
            dictionary[f"|Z| (ohm) - {label}"] = series["magnitude"]
            dictionary[f"-phi (°) - {label}"] = series["phase"]
    elif type(plot) is Nyquist:
        for series in plot.get_series():
            label = "Data"
            if context != Context.PLOTTING_TAB:
                if series.get("simulation", False):
                    label = "Sim."
                elif series.get("fit", False):
                    label = "Fit"
            else:
                label = series.get("label") or ""
            if series.get("line", False):
                label += " (line)"
            else:
                label += " (scatter)"
            key = f"Zre (ohm) - {label}"
            while key in dictionary:
                label += " "
                key = f"Zre (ohm) - {label}"
            dictionary[key] = series["real"]
            dictionary[f"-Zim (ohm) - {label}"] = series["imaginary"]
    elif type(plot) is BodeMagnitude:
        for series in plot.get_series():
            label = "Data"
            if context != Context.PLOTTING_TAB:
                if series.get("simulation", False):
                    label = "Sim."
                elif series.get("fit", False):
                    label = "Fit"
            else:
                label = series.get("label") or ""
            if series.get("line", False):
                label += " (line)"
            else:
                label += " (scatter)"
            key = f"f (Hz) - {label}"
            while key in dictionary:
                label += " "
                key = f"f (Hz) - {label}"
            dictionary[key] = series["frequency"]
            dictionary[f"|Z| (ohm) - {label}"] = series["magnitude"]
    elif type(plot) is BodePhase:
        for series in plot.get_series():
            label = "Data"
            if context != Context.PLOTTING_TAB:
                if series.get("simulation", False):
                    label = "Sim."
                elif series.get("fit", False):
                    label = "Fit"
            else:
                label = series.get("label") or ""
            if series.get("line", False):
                label += " (line)"
            else:
                label += " (scatter)"
            key = f"f (Hz) - {label}"
            while key in dictionary:
                label += " "
                key = f"f (Hz) - {label}"
            dictionary[key] = series["frequency"]
            dictionary[f"-phi (°) - {label}"] = series["phase"]
    elif type(plot) is Residuals:
        for series in plot.get_series():
            dictionary["f (Hz)"] = series["frequency"]
            dictionary["real error (%)"] = series["real"]
            dictionary["imaginary error (%)"] = series["imaginary"]
    elif type(plot) is DRT:
        for series in plot.get_series():
            label = "Data"
            if context != Context.PLOTTING_TAB:
                if series.get("simulation", False):
                    label = "Sim."
                elif series.get("fit", False):
                    label = "Fit"
            else:
                label = series.get("label") or ""
            key = f"tau (s) - {label}"
            while key in dictionary:
                label += " "
                key = f"tau (s) - {label}"
            dictionary[key] = series["tau"]
            if "gamma" in series:
                dictionary[f"gamma (ohm) - {label}"] = series["gamma"]
            elif "imaginary" in series:
                if f"gamma (ohm) - {label}" in dictionary:
                    dictionary[f"gamma, real (ohm) - {label}"] = dictionary[
                        "gamma (ohm) - {label}"
                    ]
                    del dictionary[f"gamma (ohm) - {label}"]
                dictionary[f"gamma, imag. (ohm) - {label}"] = series["imaginary"]
            elif "mean" in series:
                dictionary[f"gamma, mean (ohm) - {label}"] = series["mean"]
            elif "lower" in series and "upper" in series:
                dictionary[f"gamma, lower bound (ohm) - {label}"] = series["lower"]
                dictionary[f"gamma, upper bound (ohm) - {label}"] = series["lower"]
    elif type(plot) is Impedance:
        for series in plot.get_series():
            label = "Data"
            if context != Context.PLOTTING_TAB:
                if series.get("simulation", False):
                    label = "Sim."
                elif series.get("fit", False):
                    label = "Fit"
            else:
                label = series.get("label") or ""
            key = f"f (Hz) - {label}"
            while key in dictionary:
                label += " "
                key = f"f (Hz) - {label}"
            dictionary[key] = series["frequency"]
            dictionary["Zre (ohm)"] = series["real"]
            dictionary["-Zim (ohm)"] = series["imaginary"]
    elif type(plot) is ImpedanceReal or type(plot) is ImpedanceImaginary:
        for series in plot.get_series():
            label = "Data"
            if context != Context.PLOTTING_TAB:
                if series.get("simulation", False):
                    label = "Sim."
                elif series.get("fit", False):
                    label = "Fit"
            else:
                label = series.get("label") or ""
            key = f"f (Hz) - {label}"
            while key in dictionary:
                label += " "
                key = f"f (Hz) - {label}"
            dictionary[key] = series["x"]
            if type(plot) is ImpedanceReal:
                dictionary[f"Zre (ohm) - {label}"] = series["y"]
            else:
                dictionary[f"-Zim (ohm) - {label}"] = series["y"]
    padded_dictionary: Optional[dict] = pad_dataframe_dictionary(dictionary)
    if padded_dictionary is None:
        dpg.set_clipboard_text("")
    else:
        dpg.set_clipboard_text(
            DataFrame.from_dict(padded_dictionary).to_csv(index=False)
        )


def show_help_about(*args, **kwargs):
    x: int
    y: int
    w: int
    h: int
    x, y, w, h = calculate_window_position_dimensions(270, 100)
    window: int = dpg.generate_uuid()
    key_handler: int = dpg.generate_uuid()

    def close_window():
        if dpg.does_item_exist(window):
            dpg.delete_item(window)
        if dpg.does_item_exist(key_handler):
            dpg.delete_item(key_handler)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)

    with dpg.handler_registry(tag=key_handler):
        dpg.add_key_release_handler(
            key=dpg.mvKey_Escape,
            callback=close_window,
        )

    with dpg.window(
        label="About",
        modal=True,
        pos=(
            x,
            y,
        ),
        width=w,
        height=h,
        no_resize=True,
        on_close=close_window,
        tag=window,
    ):
        dpg.add_text(f"DearEIS ({PACKAGE_VERSION})")
        url: str
        for url in [
            "https://vyrjana.github.io/DearEIS",
            "https://github.com/vyrjana/DearEIS",
        ]:
            dpg.bind_item_theme(
                dpg.add_button(
                    label=url,
                    callback=lambda s, a, u: webbrowser.open(u),
                    user_data=url,
                    width=-1,
                ),
                themes.url_theme,
            )
    signals.emit(Signal.BLOCK_KEYBINDINGS, window=window, window_object=None)


def restore_unsaved_project_snapshots():
    parsing_errors: Dict[str, str] = {}
    unsaved_project_snapshots: List[str] = STATE.get_unsaved_project_snapshots()
    path: str
    for path in unsaved_project_snapshots:
        try:
            fp: IO
            with open(path, "r") as fp:
                project: Project = Project.from_dict(load_json(fp))
        except Exception:
            parsing_errors[path] = format_exc()
            continue
        project_tab, _ = STATE.add_project(project)
        STATE.program_window.select_tab(project_tab)
        signals.emit(Signal.SELECT_PROJECT_TAB, uuid=project.uuid)
        STATE.snapshot_project_state(project)
        signals.emit(
            Signal.RESTORE_PROJECT_STATE,
            project=project,
            project_tab=project_tab,
            state_snapshot=dump_json(project.to_dict(session=True)),
        )
    for path in unsaved_project_snapshots:
        if path not in parsing_errors:
            remove(path)
    if parsing_errors:
        total_traceback: str = ""
        traceback: str
        for path, traceback in parsing_errors.items():
            total_traceback += f"{traceback}\nThe exception above was encountered while parsing '{path}'.\n\n"
        signals.emit(
            Signal.SHOW_ERROR_MESSAGE,
            traceback=total_traceback.strip(),
            message="""
Encountered error(s) while parsing project file(s). The file(s) might be malformed, corrupted, or simply a newer version than is supported by this version of DearEIS.
            """.strip(),
        )


def initialize_program(args: Namespace):
    assert type(args) is Namespace
    signals.register(Signal.VIEWPORT_RESIZED, viewport_resized)
    dpg.set_primary_window(STATE.program_window.window, True)
    dpg.set_viewport_resize_callback(
        lambda: signals.emit(
            Signal.VIEWPORT_RESIZED,
            width=dpg.get_viewport_width(),
            height=dpg.get_viewport_height(),
        )
    )
    signals.register(
        Signal.BLOCK_KEYBINDINGS,
        lambda *a, **k: STATE.keybinding_handler.block(),
    )
    signals.register(Signal.BLOCK_KEYBINDINGS, STATE.set_active_modal_window)
    signals.register(Signal.UNBLOCK_KEYBINDINGS, STATE.keybinding_handler.unblock)
    # Signals that should cause program/project keybindings to be ignored
    # (e.g. when modal windows are shown).
    signals.register(
        Signal.SHOW_BUSY_MESSAGE,
        lambda *a, **k: signals.emit(
            Signal.BLOCK_KEYBINDINGS,
            window=STATE.program_window.busy_message.window,
            window_object=None,
        ),
    )
    signals.register(
        Signal.HIDE_BUSY_MESSAGE,
        lambda: signals.emit(Signal.UNBLOCK_KEYBINDINGS),
    )
    signals.register(
        Signal.SHOW_ERROR_MESSAGE,
        lambda *a, **k: signals.emit(
            Signal.BLOCK_KEYBINDINGS,
            window=STATE.program_window.error_message.window,
            window_object=None,
        ),
    )
    # Signals for showing/hiding the modal windows for error messages and for indicating when the
    # program is busy e.g. performing a fit.
    signals.register(Signal.SHOW_BUSY_MESSAGE, STATE.program_window.busy_message.show)
    signals.register(Signal.HIDE_BUSY_MESSAGE, STATE.program_window.busy_message.hide)
    signals.register(
        Signal.SHOW_ERROR_MESSAGE,
        lambda *a, **k: STATE.program_window.busy_message.hide(),
    )
    signals.register(Signal.SHOW_ERROR_MESSAGE, STATE.program_window.error_message.show)
    signals.register(Signal.SHOW_COMMAND_PALETTE, STATE.show_command_palette)
    # Settings and help windows
    signals.register(Signal.SHOW_HELP_ABOUT, show_help_about)
    signals.register(Signal.SHOW_HELP_LICENSES, show_license_window)
    signals.register(Signal.SHOW_SETTINGS_APPEARANCE, lambda: AppearanceSettings())
    signals.register(
        Signal.SHOW_SETTINGS_DEFAULTS, lambda: show_defaults_settings_window(STATE)
    )
    signals.register(
        Signal.SHOW_SETTINGS_KEYBINDINGS, lambda: KeybindingRemapping(STATE)
    )
    # Home tab state
    signals.register(Signal.SELECT_HOME_TAB, select_home_tab)
    STATE.set_recent_projects(paths=STATE.get_recent_projects())
    # Plots
    signals.register(Signal.SHOW_ENLARGED_PLOT, show_enlarged_plot)
    signals.register(Signal.COPY_PLOT_DATA, copy_plot_data)
    # Signals for dealing with projects and project files.
    signals.register(Signal.NEW_PROJECT, new_project)
    signals.register(Signal.SELECT_PROJECT_FILES, select_project_files)
    signals.register(Signal.LOAD_PROJECT_FILES, load_project_files)
    signals.register(Signal.RESTORE_PROJECT_STATE, restore_project_state)
    signals.register(Signal.RENAME_PROJECT, rename_project)
    signals.register(Signal.CLOSE_PROJECT, close_project)
    signals.register(Signal.SAVE_PROJECT, save_project)
    signals.register(Signal.SAVE_PROJECT_AS, save_project_as)
    signals.register(Signal.SELECT_PROJECT_TAB, select_project_tab)
    signals.register(Signal.CREATE_PROJECT_SNAPSHOT, create_project_snapshot)
    signals.register(Signal.UNDO_PROJECT_ACTION, undo)
    signals.register(Signal.REDO_PROJECT_ACTION, redo)
    signals.register(Signal.MODIFY_PROJECT_NOTES, modify_project_notes)
    signals.register(Signal.CLEAR_RECENT_PROJECTS, STATE.clear_recent_projects)
    # Signals for the data sets tab
    signals.register(Signal.LOAD_DATA_SET_FILES, load_data_set_files)
    signals.register(Signal.SELECT_DATA_SET, select_data_set)
    signals.register(Signal.SELECT_DATA_SET_FILES, select_data_set_files)
    signals.register(Signal.RENAME_DATA_SET, rename_data_set)
    signals.register(Signal.MODIFY_DATA_SET_PATH, modify_data_set_path)
    signals.register(Signal.DELETE_DATA_SET, delete_data_set)
    signals.register(Signal.SELECT_DATA_SETS_TO_AVERAGE, select_data_sets_to_average)
    signals.register(Signal.SELECT_DATA_POINTS_TO_TOGGLE, select_data_points_to_toggle)
    signals.register(Signal.SELECT_DATA_SET_MASK_TO_COPY, select_data_set_mask_to_copy)
    signals.register(Signal.SELECT_IMPEDANCE_TO_SUBTRACT, select_impedance_to_subtract)
    signals.register(Signal.TOGGLE_DATA_POINT, toggle_data_point)
    signals.register(Signal.APPLY_DATA_SET_MASK, apply_data_set_mask)
    signals.register(Signal.LOAD_SIMULATION_AS_DATA_SET, load_simulation_as_data_set)
    # Signals for the Kramers-Kronig tab
    signals.register(Signal.PERFORM_TEST, perform_test)
    signals.register(Signal.SELECT_TEST_RESULT, select_test_result)
    signals.register(Signal.DELETE_TEST_RESULT, delete_test_result)
    signals.register(Signal.APPLY_TEST_SETTINGS, apply_test_settings)
    # Signals for the DRT tab
    signals.register(Signal.PERFORM_DRT, perform_drt)
    signals.register(Signal.SELECT_DRT_RESULT, select_drt_result)
    signals.register(Signal.DELETE_DRT_RESULT, delete_drt_result)
    signals.register(Signal.APPLY_DRT_SETTINGS, apply_drt_settings)
    # Signals for the fitting tab
    signals.register(Signal.PERFORM_FIT, perform_fit)
    signals.register(Signal.SELECT_FIT_RESULT, select_fit_result)
    signals.register(Signal.COPY_OUTPUT, copy_output)
    signals.register(Signal.DELETE_FIT_RESULT, delete_fit_result)
    signals.register(Signal.APPLY_FIT_SETTINGS, apply_fit_settings)
    # Signals for the simulation tab
    signals.register(Signal.PERFORM_SIMULATION, perform_simulation)
    signals.register(Signal.SELECT_SIMULATION_RESULT, select_simulation_result)
    signals.register(Signal.DELETE_SIMULATION_RESULT, delete_simulation_result)
    signals.register(Signal.APPLY_SIMULATION_SETTINGS, apply_simulation_settings)
    # Signals for the plotting tab
    signals.register(Signal.NEW_PLOT_SETTINGS, new_plot_settings)
    signals.register(Signal.SELECT_PLOT_SETTINGS, select_plot_settings)
    signals.register(Signal.SELECT_PLOT_TYPE, select_plot_type)
    signals.register(Signal.DELETE_PLOT_SETTINGS, delete_plot_settings)
    signals.register(Signal.TOGGLE_PLOT_SERIES, toggle_plot_series)
    signals.register(Signal.RENAME_PLOT_SETTINGS, rename_plot_settings)
    signals.register(Signal.RENAME_PLOT_SERIES, rename_plot_series)
    signals.register(Signal.REORDER_PLOT_SERIES, reorder_plot_series)
    signals.register(Signal.MODIFY_PLOT_SERIES_THEME, modify_plot_series_theme)
    signals.register(
        Signal.SELECT_PLOT_APPEARANCE_SETTINGS, select_plot_appearance_settings
    )
    signals.register(
        Signal.COPY_PLOT_APPEARANCE_SETTINGS, copy_plot_appearance_settings
    )
    signals.register(Signal.EXPORT_PLOT, export_plot)
    signals.register(Signal.SAVE_PLOT, save_plot)
    signals.register(Signal.CHECK_UPDATES, perform_update_check)
    signals.register(Signal.SHOW_CHANGELOG, show_changelog)
    dpg.split_frame(delay=100)
    STATE.program_window.busy_message.resize(
        dpg.get_viewport_width(),
        dpg.get_viewport_height(),
    )
    signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Rendering assets...")
    signals.emit(Signal.RENDER_MATH)
    signals.emit(Signal.HIDE_BUSY_MESSAGE)
    # signals.register(Signal., )
    if args.data_files:
        signals.emit(Signal.NEW_PROJECT, data=args.data_files)
    if args.project_files:
        signals.emit(Signal.LOAD_PROJECT_FILES, paths=args.project_files)
    restore_unsaved_project_snapshots()
    signals.emit_backlog()
    STATE.check_version()


def program_closing():
    dirty_projects: List[Project] = list(
        filter(
            lambda _: STATE.is_project_dirty(_)
            and STATE.get_project_tab(_) is not None,
            STATE.projects,
        )
    )
    STATE.serialize_project_snapshots(dirty_projects)
    clean_projects: List[Project] = list(
        filter(
            lambda _: not STATE.is_project_dirty(_)
            and STATE.get_project_tab(_) is not None,
            STATE.projects,
        )
    )
    STATE.clear_project_backups(clean_projects)
    STATE.save_recent_projects()
    STATE.config.save()


def main():
    set_start_method("spawn")
    args: Namespace = parse_arguments()
    dpg.create_viewport(title="DearEIS")
    dpg.set_viewport_min_width(800)
    dpg.set_viewport_min_height(600)
    dpg.setup_dearpygui()
    dpg.show_viewport()
    # dpg.show_style_editor()
    # dpg.show_item_registry()
    try:
        dpg.set_frame_callback(1, lambda: initialize_program(args))
        dpg.set_exit_callback(program_closing)
        dpg.start_dearpygui()
    except Exception:
        print(format_exc())
    finally:
        dpg.destroy_context()


def debug():
    # This function is called by one of the entry points defined in setup.py.
    print("Enabling debugging features...")
    signals.DEBUG = True
    main()


if __name__ == "__main__":
    main()



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.


from time import time as _time
from uuid import uuid4 as _uuid4
import pyimpspec as _pyimpspec
from deareis.data import (
    SimulationResult,
    SimulationSettings,
)


def simulate_spectrum(
    settings: SimulationSettings,
) -> SimulationResult:
    """
    Wrapper for the `pyimpspec.simulate_spectrum` function.

    Simulate the impedance spectrum generated by a circuit in a certain frequency range.

    Parameters
    ----------
    settings: SimulationSettings
        The settings to use when performing the simulation.

    Returns
    -------
    SimulationResult
    """
    assert type(settings) is SimulationSettings, settings
    circuit: _pyimpspec.Circuit = _pyimpspec.parse_cdc(settings.cdc)
    _pyimpspec.analysis.fitting.validate_circuit(circuit)
    return SimulationResult(
        _uuid4().hex,
        _time(),
        circuit,
        settings,
    )
